"""
Time zone information
"""

tz_str = '''-1200 Y
-1100 X NUT SST
-1000 W CKT HAST HST TAHT TKT
-0900 V AKST GAMT GIT HADT HNY
-0800 U AKDT CIST HAY HNP PST PT
-0700 T HAP HNR MST PDT
-0600 S CST EAST GALT HAR HNC MDT
-0500 R CDT COT EASST ECT EST ET HAC HNE PET
-0400 Q AST BOT CLT COST EDT FKT GYT HAE HNA PYT
-0300 P ADT ART BRT CLST FKST GFT HAA PMST PYST SRT UYT WGT
-0200 O BRST FNT PMDT UYST WGST
-0100 N AZOT CVT EGT
+0000 Z EGST GMT UTC WET WT
+0100 A CET DFT WAT WEDT WEST
+0200 B CAT CEDT CEST EET SAST WAST
+0300 C EAT EEDT EEST IDT MSK
+0400 D AMT AZT GET GST KUYT MSD MUT RET SAMT SCT
+0500 E AMST AQTT AZST HMT MAWT MVT PKT TFT TJT TMT UZT YEKT
+0600 F ALMT BIOT BTT IOT KGT NOVT OMST YEKST
+0700 G CXT DAVT HOVT ICT KRAT NOVST OMSST THA WIB
+0800 H ACT AWST BDT BNT CAST HKT IRKT KRAST MYT PHT SGT ULAT WITA WST
+0900 I AWDT IRKST JST KST PWT TLT WDT WIT YAKT
+1000 K AEST ChST PGT VLAT YAKST YAPT
+1100 L AEDT LHDT MAGT NCT PONT SBT VLAST VUT
+1200 M ANAST ANAT FJT GILT MAGST MHT NZST PETST PETT TVT WFT
+1330 FJST NZDT
+1130 NFT
+1030 ACDT LHST
+0930 ACST
+0630 CCT MMT
+0545 NPT
+0530 SLT
+0430 AFT IRDT
+0330 IRST
-0230 HAT NDT
-0330 HNT NST NT
-0430 HLV VET
-0930 MART MIT'''

tz = {}
for tz_descr in map(str.split, tz_str.split('\n')):
    tz_offset = tz_descr[0]
    for tz_code in tz_descr[1:]:
        tz[tz_code] = tz_offset



"""
This module is a wrapper for the streamstats API (`streamstats documentation`_).

.. _streamstats documentation: https://streamstats.usgs.gov/streamstatsservices/#/

"""

import json

import requests


def download_workspace(workspaceID, format=''):
    """Function to download streamstats workspace.

    Parameters
    ----------
    workspaceID: string
        Service workspace received from watershed result

    format: string
        Download return format. Default will return ESRI geodatabase zipfile.
        'SHAPE' will return a zip file containing shape format.

    Returns
    -------
    r: geodatabase or shapefiles
        A zip file containing the workspace contents, in either a
        geodatabase or shape files.

    """
    payload = {'workspaceID': workspaceID, 'format': format}
    url = 'https://streamstats.usgs.gov/streamstatsservices/download'

    r = requests.get(url, params=payload)

    r.raise_for_status()
    return r
    # data = r.raw.read()

    # with open(filepath, 'wb') as f:
    #    f.write(data)

    # return


def get_sample_watershed():
    """Sample function to get a watershed object for a location in NY.

    Makes the function call :obj:`dataretrieval.streamstats.get_watershed`
    with the parameters 'NY', -74.524, 43.939, and returns the watershed
    object.

    Returns
    -------
    Watershed: :obj:`dataretrieval.streamstats.Watershed`
        Custom object that contains the watershed information as extracted
        from the streamstats JSON object.

    """
    return get_watershed('NY', -74.524, 43.939)


def get_watershed(
    rcode,
    xlocation,
    ylocation,
    crs=4326,
    includeparameters=True,
    includeflowtypes=False,
    includefeatures=True,
    simplify=True,
    format='geojson',
):
    """Get watershed object based on location

    **Streamstats documentation:**
    Returns a watershed object. The request configuration will determine the
    overall request response. However all returns will return a watershed
    object with at least the workspaceid. The workspace id is the id to the
    service workspace where files are stored and can be used for further
    processing such as for downloads and flow statistic computations.

    See: https://streamstats.usgs.gov/streamstatsservices/#/ for more
    information.

    Parameters
    ----------
    rcode: string
        StreamStats 2-3 character code that identifies the Study Area --
        either a State or a Regional Study.
    xlocation: float
        X location of the most downstream point of desired study area.
    ylocation: float
        Y location of the most downstream point of desired study area.
    crs: integer, string, optional
        ESPSG spatial reference code, default is 4326
    includeparameters: bool, optional
        Boolean flag to include parameters in response.
    includeflowtypes: bool, string, optional
        Not yet implemented. Would be a comma separated list of region flow
        types to compute with the default being True
    includefeatures: list, optional
        Comma separated list of features to include in response.
    simplify: bool, optional
        Boolean flag controlling whether or not to simplify the returned
        result.

    Returns
    -------
    Watershed: :obj:`dataretrieval.streamstats.Watershed`
        Custom object that contains the watershed information as extracted
        from the streamstats JSON object.

    """
    payload = {
        'rcode': rcode,
        'xlocation': xlocation,
        'ylocation': ylocation,
        'crs': crs,
        'includeparameters': includeparameters,
        'includeflowtypes': includeflowtypes,
        'includefeatures': includefeatures,
        'simplify': simplify,
    }
    url = 'https://streamstats.usgs.gov/streamstatsservices/watershed.geojson'

    r = requests.get(url, params=payload)

    r.raise_for_status()

    if format == 'geojson':
        return r

    if format == 'shape':
        # use Fiona to return a shape object
        pass

    if format == 'object':
        # return a python object
        pass

    data = json.loads(r.text)
    return Watershed.from_streamstats_json(data)


class Watershed:
    """Class to extract information from the streamstats JSON object."""

    @classmethod
    def from_streamstats_json(cls, streamstats_json):
        """Method that creates a Watershed object from a streamstats JSON."""
        cls.watershed_point = streamstats_json['featurecollection'][0]['feature']
        cls.watershed_polygon = streamstats_json['featurecollection'][1]['feature']
        cls.parameters = streamstats_json['parameters']
        cls._workspaceID = streamstats_json['workspaceID']
        return cls

    def __init__(self, rcode, xlocation, ylocation):
        """Init method that calls the :obj:`from_streamstats_json` method."""
        self = get_watershed(rcode, xlocation, ylocation)



"""
Tool for downloading data from the Water Quality Portal (https://waterqualitydata.us)

See https://waterqualitydata.us/webservices_documentation for API reference

.. todo::

    - implement other services like Organization, Activity, etc.

"""
import warnings
from io import StringIO

import pandas as pd

from .utils import BaseMetadata, query


def get_results(ssl_check=True, **kwargs):
    """Query the WQP for results.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    siteid: string
        Concatenate an agency code, a hyphen ("-"), and a site-identification
        number.
    statecode: string
        Concatenate 'US', a colon (":"), and a FIPS numeric code
        (Example: Illinois is US:17)
    countycode: string
        A FIPS county code
    huc: string
        One or more eight-digit hydrologic units, delimited by semicolons.
    bBox: string
        Bounding box (Example: bBox=-92.8,44.2,-88.9,46.0)
    lat: string
        Latitude for radial search, expressed in decimal degrees, WGS84
    long: string
        Longitude for radial search
    within: string
        Distance for a radial search, expressed in decimal miles
    pCode: string
        One or more five-digit USGS parameter codes, separated by semicolons.
        NWIS only.
    startDateLo: string
        Date of earliest desired data-collection activity,
        expressed as 'MM-DD-YYYY'
    startDateHi: string
        Date of last desired data-collection activity,
        expressed as 'MM-DD-YYYY'
    characteristicName: string
        One or more case-sensitive characteristic names, separated by
        semicolons. (See https://www.waterqualitydata.us/public_srsnames/
        for available characteristic names)
    mimeType: string
        String specifying the output format which is 'csv' by default but can
        be 'geojson'
    zip: string
        Parameter to stream compressed data, if 'yes', or uncompressed data
        if 'no'. Default is 'no'.

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom ``dataretrieval`` metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get results within a radial distance of a point
        >>> df, md = dataretrieval.wqp.get_results(
        ...     lat='44.2', long='-88.9', within='0.5'
        ... )

        >>> # Get results within a bounding box
        >>> df, md = dataretrieval.wqp.get_results(bBox='-92.8,44.2,-88.9,46.0')

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)
    response = query(wqp_url('Result'), kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')
    return df, WQP_Metadata(response)


def what_sites(ssl_check=True, **kwargs):
    """Search WQP for sites within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get sites within a radial distance of a point
        >>> df, md = dataretrieval.wqp.what_sites(
        ...     lat='44.2', long='-88.9', within='2.5'
        ... )

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('Station')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_organizations(ssl_check=True, **kwargs):
    """Search WQP for organizations within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get all organizations in the WQP
        >>> df, md = dataretrieval.wqp.what_organizations()

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('Organization')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_projects(ssl_check=True, **kwargs):
    """Search WQP for projects within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get projects within a HUC region
        >>> df, md = dataretrieval.wqp.what_projects(huc='19')

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('Project')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_activities(ssl_check=True, **kwargs):
    """Search WQP for activities within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get activities within Washington D.C.
        >>> # during a specific time period
        >>> df, md = dataretrieval.wqp.what_activities(
        ...     statecode='US:11', startDateLo='12-30-2019', startDateHi='01-01-2020'
        ... )

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('Activity')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_detection_limits(ssl_check=True, **kwargs):
    """Search WQP for result detection limits within a region with specific
    data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get detection limits for Nitrite measurements in Rhode Island
        >>> # between specific dates
        >>> df, md = dataretrieval.wqp.what_detection_limits(
        ...     statecode='US:44',
        ...     characteristicName='Nitrite',
        ...     startDateLo='01-01-2021',
        ...     startDateHi='02-20-2021',
        ... )

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('ResultDetectionQuantitationLimit')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_habitat_metrics(ssl_check=True, **kwargs):
    """Search WQP for habitat metrics within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get habitat metrics for a state (Rhode Island in this case)
        >>> df, md = dataretrieval.wqp.what_habitat_metrics(statecode='US:44')

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('BiologicalMetric')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_project_weights(ssl_check=True, **kwargs):
    """Search WQP for project weights within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get project weights for a state (North Dakota in this case)
        >>> # within a set time period
        >>> df, md = dataretrieval.wqp.what_project_weights(
        ...     statecode='US:38', startDateLo='01-01-2006', startDateHi='01-01-2009'
        ... )

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('ProjectMonitoringLocationWeighting')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def what_activity_metrics(ssl_check=True, **kwargs):
    """Search WQP for activity metrics within a region with specific data.

    Parameters
    ----------
    ssl_check: bool
        If True, check the SSL certificate. Default is True. If False, SSL
        certificate is not checked.
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.wqp.get_results`

    Returns
    -------
    df: ``pandas.DataFrame``
        Formatted data returned from the API query.
    md: :obj:`dataretrieval.utils.Metadata`
        Custom metadata object pertaining to the query.

    Examples
    --------
    .. code::

        >>> # Get activity metrics for a state (North Dakota in this case)
        >>> # within a set time period
        >>> df, md = dataretrieval.wqp.what_activity_metrics(
        ...     statecode='US:38', startDateLo='07-01-2006', startDateHi='12-01-2006'
        ... )

    """
    _warn_v3_profiles_outage()

    kwargs = _alter_kwargs(kwargs)

    url = wqp_url('ActivityMetric')
    response = query(url, payload=kwargs, delimiter=';', ssl_check=ssl_check)

    df = pd.read_csv(StringIO(response.text), delimiter=',')

    return df, WQP_Metadata(response)


def wqp_url(service):
    """Construct the WQP URL for a given service."""
    base_url = 'https://www.waterqualitydata.us/data/'
    return f'{base_url}{service}/Search?'


class WQP_Metadata(BaseMetadata):
    """Metadata class for WQP service, derived from BaseMetadata.

    Attributes
    ----------
    url : str
        Response url
    query_time: datetme.timedelta
        Response elapsed time
    header: requests.structures.CaseInsensitiveDict
        Response headers
    comments: None
        Metadata comments. WQP does not return comments.
    site_info: tuple[pd.DataFrame, NWIS_Metadata] | None
        Site information if the query included `sites`, `site` or `site_no`.
    """

    def __init__(self, response, **parameters) -> None:
        """Generates a standard set of metadata informed by the response with specific
        metadata for WQP data.

        Parameters
        ----------
        response: Response
            Response object from requests module

        parameters: unpacked dictionary
            Unpacked dictionary of the parameters supplied in the request

        Returns
        -------
        md: :obj:`dataretrieval.wqp.WQP_Metadata`
            A ``dataretrieval`` custom :obj:`dataretrieval.wqp.WQP_Metadata` object.

        """

        super().__init__(response)

        self._parameters = parameters

        @property
        def site_info(self):
            if 'sites' in self._parameters:
                return what_sites(sites=parameters['sites'])
            elif 'site' in self._parameters:
                return what_sites(sites=parameters['site'])
            elif 'site_no' in self._parameters:
                return what_sites(sites=parameters['site_no'])


def _alter_kwargs(kwargs):
    """Private function to manipulate **kwargs.

    Not all query parameters are currently supported by ``dataretrieval``,
    so this function is used to set some of them and raise warnings to the
    user so they are aware of which are being hard-set.

    """
    if kwargs.get('zip', 'no') == 'yes':
        warnings.warn('Compressed data not yet supported, zip set to no.')
    kwargs['zip'] = 'no'

    if kwargs.get('mimeType', 'csv') == 'geojson':
        warnings.warn('GeoJSON not yet supported, mimeType set to csv.')
    kwargs['mimeType'] = 'csv'

    return kwargs

def _warn_v3_profiles_outage():
    """Private function for warning message about WQX 3.0 profiles
    """

    warnings.warn(('USGS discrete water quality data availability '
                   'and format are changing. Beginning in March 2024 '
                   'the data obtained from legacy profiles will not '
                   'include new USGS data or recent updates to existing '
                   'data. To view the status of changes in data '
                   'availability and code functionality, visit: '
                   'https://doi-usgs.github.io/dataRetrieval/articles/Status.html. '
                   'If you have additional questions about these changes, '
                   'email CompTools@usgs.gov.'))



"""
Tools for retrieving data from the National Atmospheric Deposition Program
(NADP) including the National Trends Network (NTN), the Mercury Deposition
Network (MDN).

National Trends Network
-----------------------
The  NTN provides long-term records of precipitation chemistry across the
United States. See https://nadp.slh.wisc.edu/ntn for more info.

Mercury Deposition Network
--------------------------
The MDN provides long-term records of total mercury (Hg) concentration and
deposition in precipitation in the United States and Canada. For more
information visit https://nadp.slh.wisc.edu/networks/mercury-deposition-network/

Notes
-----
Gridded data on NADP is served as zipped tif files. Functions in this module
will either download and extract the data, when a path is specified, or open
the data as a GDAL memory-mapped file when no path is specified.

.. todo::

    - include AIRMoN, AMNet, and AMoN
    - flexible handling of strings for parameters and measurement types
    - add errorchecking
    - add tests

"""

import io
import os
import re
import zipfile
from os.path import basename

import requests

NADP_URL = 'https://nadp.slh.wisc.edu'
NADP_MAP_EXT = 'filelib/maps'

NTN_CONC_PARAMS = ['pH', 'So4', 'NO3', 'NH4', 'Ca', 'Mg', 'K', 'Na', 'Cl', 'Br']
NTN_DEP_PARAMS = [
    'H',
    'So4',
    'NO3',
    'NH4',
    'Ca',
    'Mg',
    'K',
    'Na',
    'Cl',
    'Br',
    'N',
    'SPlusN',
]

NTN_MEAS_TYPE = ['conc', 'dep', 'precip']  # concentration or deposition


class NADP_ZipFile(zipfile.ZipFile):
    """Extend zipfile.ZipFile for working on data from NADP"""

    def tif_name(self):
        """Get the name of the tif file in the zip file."""
        filenames = self.namelist()
        r = re.compile('.*tif$')
        tif_list = list(filter(r.match, filenames))
        return tif_list[0]

    def tif(self):
        """Read the tif file in the zip file."""
        return self.read(self.tif_name())


def get_annual_MDN_map(measurement_type, year, path):
    """Download a MDN map from NDAP.

    This function looks for a zip file containing gridded information at:
    https://nadp.slh.wisc.edu/maps-data/mdn-gradient-maps/.
    The function will download the zip file and extract it, exposing the tif
    file if a path is provided.

    Parameters
    ----------
    measurement_type: string
        The type of measurement (concentration or deposition) as a string,
        either 'conc' or 'dep' respectively.

    year: string
        Year as a string 'YYYY'

    path: string
        Download directory.

    Returns
    -------
    path: string
        Path that zip file was extracted into if path was specified.

    Examples
    --------
    .. code::

        >>> # get map of mercury concentration in 2010 and extract it to a path
        >>> data_path = dataretrieval.nadp.get_annual_MDN_map(
        ...     measurement_type='conc', year='2010', path='somepath'
        ... )

    """
    url = f'{NADP_URL}/{NADP_MAP_EXT}/MDN/grids/'

    filename = f'Hg_{measurement_type}_{year}.zip'

    z = get_zip(url, filename)

    if path:
        z.extractall(path)

    return f'{path}{os.sep}{basename(filename)}'


def get_annual_NTN_map(measurement_type, measurement=None, year=None, path='.'):
    """Download a NTN map from NDAP.

    This function looks for a zip file containing gridded information at:
    https://nadp.slh.wisc.edu/maps-data/ntn-gradient-maps/.
    The function will download the zip file and extract it, exposing the tif
    file at the provided path.

    .. note::

        Measurement type abbreviations for concentration and deposition are
        all lower-case, but for precipitation data, the first letter must be
        capitalized!

    Parameters
    ----------
    measurement : string
        The measured constituent to return.
    measurement_type : string
        The type of measurement, 'conc', 'dep', or 'Precip', which represent
        concentration, deposition, or precipitation respectively.
    year : string
        Year as a string 'YYYY'
    path : string
        Download directory, defaults to current directory if not specified.

    Returns
    -------
    path: string
        Path that zip file was extracted into if path was specified.

    Examples
    --------
    .. code::

        >>> # get a map of precipitation in 2015 and extract it to a path
        >>> data_path = dataretrieval.nadp.get_annual_NTN_map(
        ...     measurement_type='Precip', year='2015', path='somepath'
        ... )

    """
    url = f'{NADP_URL}/{NADP_MAP_EXT}/NTN/grids/{year}/'

    filename = f'{measurement_type}_{year}.zip'

    if measurement:
        filename = f'{measurement}_{filename}'

    z = get_zip(url, filename)

    if path:
        z.extractall(path)

    return f'{path}{os.sep}{basename(filename)}'


def get_zip(url, filename):
    """Gets a ZipFile at url and returns it

    Parameters
    ----------
    url : string
        URL to zip file

    filename : string
        Name of zip file

    Returns
    -------
    ZipFile

    .. todo::

        finish docstring

    """
    req = requests.get(url + filename)
    req.raise_for_status()

    # z = zipfile.ZipFile(io.BytesIO(req.content))
    z = NADP_ZipFile(io.BytesIO(req.content))
    return z



from geopandas import GeoDataFrame

from dataretrieval.nldi import (
    NLDI_API_BASE_URL,
    get_basin,
    get_features,
    get_flowlines,
    search,
)


def mock_request_data_sources(requests_mock):
    request_url = f'{NLDI_API_BASE_URL}/'
    available_data_sources = [
        {'source': 'ca_gages'},
        {'source': 'census2020-nhdpv2'},
        {'source': 'epa_nrsa'},
        {'source': 'geoconnex-demo'},
        {'source': 'gfv11_pois'},
        {'source': 'huc12pp'},
        {'source': 'huc12pp_102020'},
        {'source': 'nmwdi-st'},
        {'source': 'npdes'},
        {'source': 'nwisgw'},
        {'source': 'nwissite'},
        {'source': 'ref_gage'},
        {'source': 'vigil'},
        {'source': 'wade'},
        {'source': 'WQP'},
        {'source': 'comid'},
    ]
    requests_mock.get(
        request_url, json=available_data_sources, headers={'mock_header': 'value'}
    )


def mock_request(requests_mock, request_url, file_path):
    with open(file_path) as text:
        requests_mock.get(
            request_url, text=text.read(), headers={'mock_header': 'value'}
        )


def test_get_basin(requests_mock):
    """Tests NLDI get basin query"""
    request_url = (
        f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/basin'
        f'?simplified=true&splitCatchment=false'
    )
    response_file_path = 'data/nldi_get_basin.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_basin(feature_source='WQP', feature_id='USGS-054279485')
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 1


def test_get_flowlines(requests_mock):
    """Tests NLDI get flowlines query using feature source as the origin"""
    request_url = (
        f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/navigation/UM/flowlines'
        f'?distance=5&trimStart=false'
    )
    response_file_path = 'data/nldi_get_flowlines.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_flowlines(
        feature_source='WQP', feature_id='USGS-054279485', navigation_mode='UM'
    )
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 2


def test_get_flowlines_by_comid(requests_mock):
    """Tests NLDI get flowlines query using comid as the origin"""
    request_url = (
        f'{NLDI_API_BASE_URL}/comid/13294314/navigation/UM/flowlines?distance=50'
    )
    response_file_path = 'data/nldi_get_flowlines_by_comid.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_flowlines(navigation_mode='UM', comid=13294314, distance=50)
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 16


def test_features_by_feature_source_with_navigation(requests_mock):
    """Tests NLDI get features query using feature source as the origin
    with navigation mode
    """
    request_url = (
        f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/navigation/UM/nwissite?distance=50'
    )
    response_file_path = 'data/nldi_get_features_by_feature_source_with_nav_mode.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_features(
        feature_source='WQP',
        feature_id='USGS-054279485',
        data_source='nwissite',
        navigation_mode='UM',
        distance=50,
    )
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 108


def test_features_by_feature_source_without_navigation(requests_mock):
    """Tests NLDI get features query using feature source as the origin
    without navigation mode
    """
    request_url = f'{NLDI_API_BASE_URL}/WQP/USGS-054279485'
    response_file_path = (
        'data/nldi_get_features_by_feature_source_without_nav_mode.json'
    )
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_features(feature_source='WQP', feature_id='USGS-054279485')
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 10


def test_get_features_by_comid(requests_mock):
    """Tests NLDI get features query using comid as the origin"""
    request_url = f'{NLDI_API_BASE_URL}/comid/13294314/navigation/UM/WQP?distance=5'
    response_file_path = 'data/nldi_get_features_by_comid.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_features(
        comid=13294314, data_source='WQP', navigation_mode='UM', distance=5
    )
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 405


def test_get_features_by_lat_long(requests_mock):
    """Tests NLDI get features query using lat/long as the origin"""
    request_url = (
        f'{NLDI_API_BASE_URL}/comid/position?coords=POINT%28-89.509%2043.087%29'
    )
    response_file_path = 'data/nldi_get_features_by_lat_long.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    gdf = get_features(lat=43.087, long=-89.509)
    assert isinstance(gdf, GeoDataFrame)
    assert gdf.size == 6


def test_search_for_basin(requests_mock):
    """Tests NLDI search query for basin"""
    request_url = f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/basin'
    response_file_path = 'data/nldi_get_basin.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(
        feature_source='WQP', feature_id='USGS-054279485', find='basin'
    )
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'Polygon'
    assert len(search_results['features'][0]['geometry']['coordinates'][0]) == 122


def test_search_for_flowlines(requests_mock):
    """Tests NLDI search query for flowlines"""
    request_url = f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/navigation/UM/flowlines'
    response_file_path = 'data/nldi_get_flowlines.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(
        feature_source='WQP',
        feature_id='USGS-054279485',
        navigation_mode='UM',
        find='flowlines',
    )
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'LineString'
    assert len(search_results['features'][0]['geometry']['coordinates']) == 27


def test_search_for_flowlines_by_comid(requests_mock):
    """Tests NLDI search query for flowlines by comid"""
    request_url = f'{NLDI_API_BASE_URL}/comid/13294314/navigation/UM/flowlines'
    response_file_path = 'data/nldi_get_flowlines_by_comid.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(comid=13294314, navigation_mode='UM', find='flowlines')
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'LineString'
    assert len(search_results['features'][0]['geometry']['coordinates']) == 27


def test_search_for_features_by_feature_source_with_navigation(requests_mock):
    """Tests NLDI search query for features by feature source"""
    request_url = (
        f'{NLDI_API_BASE_URL}/WQP/USGS-054279485/navigation/UM/nwissite?distance=50'
    )
    response_file_path = 'data/nldi_get_features_by_feature_source_with_nav_mode.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(
        feature_source='WQP',
        feature_id='USGS-054279485',
        data_source='nwissite',
        navigation_mode='UM',
        find='features',
    )
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'Point'
    assert len(search_results['features']) == 9


def test_search_for_features_by_feature_source_without_navigation(requests_mock):
    """Tests NLDI search query for features by feature source"""
    request_url = f'{NLDI_API_BASE_URL}/WQP/USGS-054279485'
    response_file_path = (
        'data/nldi_get_features_by_feature_source_without_nav_mode.json'
    )
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(
        feature_source='WQP', feature_id='USGS-054279485', find='features'
    )
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'Point'
    assert len(search_results['features']) == 1


def test_search_for_features_by_comid(requests_mock):
    """Tests NLDI search query for features by comid"""
    request_url = f'{NLDI_API_BASE_URL}/comid/13294314/navigation/UM/WQP?distance=5'
    response_file_path = 'data/nldi_get_features_by_comid.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(
        comid=13294314,
        data_source='WQP',
        navigation_mode='UM',
        find='features',
        distance=5,
    )
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'Point'
    assert len(search_results['features']) == 45


def test_search_for_features_by_lat_long(requests_mock):
    """Tests NLDI search query for features by lat/long"""
    request_url = (
        f'{NLDI_API_BASE_URL}/comid/position?coords=POINT%28-89.509%2043.087%29'
    )
    response_file_path = 'data/nldi_get_features_by_lat_long.json'
    mock_request_data_sources(requests_mock)
    mock_request(requests_mock, request_url, response_file_path)

    search_results = search(lat=43.087, long=-89.509, find='features')
    assert isinstance(search_results, dict)
    assert search_results['features'][0]['type'] == 'Feature'
    assert search_results['features'][0]['geometry']['type'] == 'LineString'
    assert len(search_results['features'][0]['geometry']['coordinates']) == 27



"""
Useful utilities for data munging.
"""
import warnings

import pandas as pd
import requests

import dataretrieval
from dataretrieval.codes import tz


def to_str(listlike, delimiter=','):
    """Translates list-like objects into strings.

    Parameters
    ----------
    listlike: list-like object
        An object that is a list, or list-like
        (e.g., ``pandas.core.series.Series``)
    delimiter: string, optional
        The delimiter that is placed between entries in listlike when it is
        turned into a string. Default value is a comma.

    Returns
    -------
    listlike: string
        The listlike object as string separated by the delimiter

    Examples
    --------
    .. doctest::

        >>> dataretrieval.utils.to_str([1, 'a', 2])
        '1,a,2'

        >>> dataretrieval.utils.to_str([0, 10, 42], delimiter='+')
        '0+10+42'

    """
    if type(listlike) == list:
        return delimiter.join([str(x) for x in listlike])

    elif type(listlike) == pd.core.series.Series:
        return delimiter.join(listlike.tolist())

    elif type(listlike) == pd.core.indexes.base.Index:
        return delimiter.join(listlike.tolist())

    elif type(listlike) == str:
        return listlike


def format_datetime(df, date_field, time_field, tz_field):
    """Creates a datetime field from separate date, time, and
    time zone fields.

    Assumes ISO 8601.

    Parameters
    ----------
    df: ``pandas.DataFrame``
        A data frame containing date, time, and timezone fields.
    date_field: string
        Name of date column in df.
    time_field: string
        Name of time column in df.
    tz_field: string
        Name of time zone column in df.

    Returns
    -------
    df: ``pandas.DataFrame``
        The data frame with a formatted 'datetime' column

    """
    # create a datetime index from the columns in qwdata response
    df[tz_field] = df[tz_field].map(tz)

    df['datetime'] = pd.to_datetime(
        df[date_field] + ' ' + df[time_field] + ' ' + df[tz_field],
        format='ISO8601',
        utc=True,
    )

    # if there are any incomplete dates, warn the user
    if df['datetime'].isna().any():
        count = df['datetime'].isna().sum()
        warnings.warn(
            f'Warning: {count} incomplete dates found, '
            + 'consider setting datetime_index to False.',
            UserWarning,
        )

    return df


class BaseMetadata:
    """Base class for metadata.

    Attributes
    ----------
    url : str
        Response url
    query_time: datetme.timedelta
        Response elapsed time
    header: requests.structures.CaseInsensitiveDict
        Response headers

    """

    def __init__(self, response) -> None:
        """Generates a standard set of metadata informed by the response.

        Parameters
        ----------
        response: Response
            Response object from requests module

        Returns
        -------
        md: :obj:`dataretrieval.utils.BaseMetadata`
            A ``dataretrieval`` custom :obj:`dataretrieval.utils.BaseMetadata` object.

        """

        # These are built from the API response
        self.url = response.url
        self.query_time = response.elapsed
        self.header = response.headers
        self.comment = None

        # # not sure what statistic_info is
        # self.statistic_info = None

        # # disclaimer seems to be only part of importWaterML1
        # self.disclaimer = None

    # These properties are to be set by `nwis` or `wqp`-specific metadata classes.
    @property
    def site_info(self):
        raise NotImplementedError(
            'site_info must be implemented by utils.BaseMetadata children'
        )

    @property
    def variable_info(self):
        raise NotImplementedError(
            'variable_info must be implemented by utils.BaseMetadata children'
        )

    def __repr__(self) -> str:
        return f'{type(self).__name__}(url={self.url})'


def query(url, payload, delimiter=',', ssl_check=True):
    """Send a query.

    Wrapper for requests.get that handles errors, converts listed
    query parameters to comma separated strings, and returns response.

    Parameters
    ----------
    url: string
        URL to query
    payload: dict
        query parameters passed to ``requests.get``
    delimiter: string
        delimiter to use with lists
    ssl_check: bool
        If True, check SSL certificates, if False, do not check SSL,
        default is True

    Returns
    -------
    string: query response
        The response from the API query ``requests.get`` function call.
    """

    for key, value in payload.items():
        payload[key] = to_str(value, delimiter)
    # for index in range(len(payload)):
    #    key, value = payload[index]
    #    payload[index] = (key, to_str(value))

    # define the user agent for the query
    user_agent = {'user-agent': f'python-dataretrieval/{dataretrieval.__version__}'}

    response = requests.get(url, params=payload, headers=user_agent, verify=ssl_check)

    if response.status_code == 400:
        raise ValueError(
            f'Bad Request, check that your parameters are correct. URL: {response.url}'
        )
    elif response.status_code == 404:
        raise ValueError(
            'Page Not Found Error. May be the result of an empty query. '
            + f'URL: {response.url}'
        )
    elif response.status_code == 414:
        _reason = response.reason
        _example = """
                    # n is the number of chunks to divide the query into \n
                    split_list = np.array_split(site_list, n)
                    data_list = []  # list to store chunk results in \n
                    # loop through chunks and make requests \n
                    for site_list in split_list: \n
                        data = nwis.get_record(sites=site_list, service='dv', \n
                                               start=start, end=end) \n
                        data_list.append(data)  # append results to list"""
        raise ValueError(
            'Request URL too long. Modify your query to use fewer sites. '
            + f'API response reason: {_reason}. Pseudo-code example of how to '
            + f'split your query: \n {_example}'
        )

    if response.text.startswith('No sites/data'):
        raise NoSitesError(response.url)

    return response


class NoSitesError(Exception):
    """Custom error class used when selection criteria returns no sites/data."""

    def __init__(self, url):
        self.url = url

    def __str__(self):
        return (
            'No sites/data found using the selection criteria specified in url: '
            '{url}'
        ).format(url=self.url)



import pytest
import requests
import datetime

from pandas import DataFrame

from dataretrieval.wqp import (get_results, what_sites, what_organizations,
                               what_projects, what_activities,
                               what_detection_limits, what_habitat_metrics,
                               what_project_weights, what_activity_metrics,
                               _alter_kwargs)


def test_get_ratings(requests_mock):
    """Tests water quality portal ratings query"""
    request_url = "https://www.waterqualitydata.us/data/Result/Search?siteid=WIDNR_WQX-10032762" \
                  "&characteristicName=Specific+conductance&startDateLo=05-01-2011&startDateHi=09-30-2011" \
                  "&zip=no&mimeType=csv"
    response_file_path = 'data/wqp_results.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_results(siteid='WIDNR_WQX-10032762',
                          characteristicName = 'Specific conductance',
                          startDateLo='05-01-2011', startDateHi='09-30-2011')
    assert type(df) is DataFrame
    assert df.size == 315
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_sites(requests_mock):
    """Tests Water quality portal sites query"""
    request_url = "https://www.waterqualitydata.us/data/Station/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_sites.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_sites(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 239868
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_organizations(requests_mock):
    """Tests Water quality portal organizations query"""
    request_url = "https://www.waterqualitydata.us/data/Organization/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_organizations.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_organizations(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 576
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_projects(requests_mock):
    """Tests Water quality portal projects query"""
    request_url = "https://www.waterqualitydata.us/data/Project/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_projects.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_projects(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 530
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_activities(requests_mock):
    """Tests Water quality portal activities query"""
    request_url = "https://www.waterqualitydata.us/data/Activity/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_activities.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_activities(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 5087443
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_detection_limits(requests_mock):
    """Tests Water quality portal detection limits query"""
    request_url = "https://www.waterqualitydata.us/data/ResultDetectionQuantitationLimit/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_detection_limits.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_detection_limits(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 98770
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_habitat_metrics(requests_mock):
    """Tests Water quality portal habitat metrics query"""
    request_url = "https://www.waterqualitydata.us/data/BiologicalMetric/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_habitat_metrics.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_habitat_metrics(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 48114
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_project_weights(requests_mock):
    """Tests Water quality portal project weights query"""
    request_url = "https://www.waterqualitydata.us/data/ProjectMonitoringLocationWeighting/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_project_weights.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_project_weights(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 33098
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def test_what_activity_metrics(requests_mock):
    """Tests Water quality portal activity metrics query"""
    request_url = "https://www.waterqualitydata.us/data/ActivityMetric/Search?statecode=US%3A34&characteristicName=Chloride&zip=no" \
                  "&mimeType=csv"
    response_file_path = 'data/wqp_activity_metrics.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = what_activity_metrics(statecode="US:34", characteristicName="Chloride")
    assert type(df) is DataFrame
    assert df.size == 378
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    assert md.comment is None


def mock_request(requests_mock, request_url, file_path):
    with open(file_path) as text:
        requests_mock.get(request_url, text=text.read(), headers={"mock_header": "value"})


class TestAlterKwargs:
    """Tests for keyword alteration.
    """
    def test_alter_kwargs_zip(self):
        """Tests that zip kwarg is altered correctly and warning is thrown."""
        kwargs = {"zip": "yes", "mimeType": "csv"}
        with pytest.warns(UserWarning):
            kwargs = _alter_kwargs(kwargs)
        assert kwargs == {"zip": "no", "mimeType": "csv"}

    def test_alter_kwargs_mimetype(self):
        """Tests that mimetype kwarg is altered correctly and warning is thrown."""
        kwargs = {"zip": "no", "mimeType": "geojson"}
        with pytest.warns(UserWarning):
            kwargs = _alter_kwargs(kwargs)
        assert kwargs == {"zip": "no", "mimeType": "csv"}


import datetime

import pytest
from pandas import DataFrame

from dataretrieval.nwis import (
    get_discharge_measurements,
    get_discharge_peaks,
    get_dv,
    get_gwlevels,
    get_info,
    get_iv,
    get_pmcodes,
    get_qwdata,
    get_ratings,
    get_record,
    get_stats,
    get_water_use,
    query_waterdata,
    query_waterservices,
    what_sites
)
from dataretrieval.utils import NoSitesError


def test_query_waterdata_validation():
    """Tests the validation parameters of the query_waterservices method"""
    with pytest.raises(TypeError) as type_error:
        query_waterdata(service='pmcodes', format='rdb')
    assert 'Query must specify a major filter: site_no, stateCd, bBox' == str(type_error.value)

    with pytest.raises(TypeError) as type_error:
        query_waterdata(service=None, site_no='sites')
    assert 'Service not recognized' == str(type_error.value)

    with pytest.raises(TypeError) as type_error:
        query_waterdata(service='pmcodes', nw_longitude_va='something')
    assert 'One or more lat/long coordinates missing or invalid.' == str(type_error.value)


def test_query_waterservices_validation():
    """Tests the validation parameters of the query_waterservices method"""
    with pytest.raises(TypeError) as type_error:
        query_waterservices(service='dv', format='rdb')
    assert 'Query must specify a major filter: sites, stateCd, bBox, huc, or countyCd' == str(type_error.value)

    with pytest.raises(TypeError) as type_error:
        query_waterservices(service=None, sites='sites')
    assert 'Service not recognized' == str(type_error.value)


def test_query_validation(requests_mock):
    request_url = "https://waterservices.usgs.gov/nwis/stat?sites=bad_site_id&format=rdb"
    requests_mock.get(request_url, status_code=400)
    with pytest.raises(ValueError) as type_error:
        get_stats(sites="bad_site_id")
    assert request_url in str(type_error)

    request_url = "https://waterservices.usgs.gov/nwis/stat?sites=123456&format=rdb"
    requests_mock.get(request_url,
                      text="No sites/data found using the selection criteria specified")
    with pytest.raises(NoSitesError) as no_sites_error:
        get_stats(sites="123456")
    assert request_url in str(no_sites_error)


def test_get_record_validation():
    """Tests the validation parameters of the get_record method"""
    with pytest.raises(TypeError) as type_error:
        get_record(sites=['01491000'], service='not_a_service')
    assert 'Unrecognized service: not_a_service' == str(type_error.value)


def test_get_dv(requests_mock):
    """Tests get_dv method correctly generates the request url and returns the result in a DataFrame"""
    format = "json"
    site = '01491000%2C01645000'
    request_url = 'https://waterservices.usgs.gov/nwis/dv?format={}' \
                  '&startDT=2020-02-14&endDT=2020-02-15&sites={}'.format(format, site)
    response_file_path = 'data/waterservices_dv.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_dv(sites=["01491000", "01645000"], start='2020-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 8
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_dv_site_value_types(requests_mock, site_input_type_list):
    """Tests get_dv method for valid input types for the 'sites' parameter"""
    _format = "json"
    site = '01491000'
    request_url = 'https://waterservices.usgs.gov/nwis/dv?format={}' \
                  '&startDT=2020-02-14&endDT=2020-02-15&sites={}'.format(_format, site)
    response_file_path = 'data/waterservices_dv.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df, md = get_dv(sites=sites, start='2020-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 8


def test_get_iv(requests_mock):
    """Tests get_iv method correctly generates the request url and returns the result in a DataFrame"""
    format = "json"
    site = '01491000%2C01645000'
    request_url = 'https://waterservices.usgs.gov/nwis/iv?format={}' \
                  '&startDT=2019-02-14&endDT=2020-02-15&sites={}'.format(format, site)
    response_file_path = 'data/waterservices_iv.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_iv(sites=["01491000", "01645000"], start='2019-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 563380
    assert md.url == request_url
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_iv_site_value_types(requests_mock, site_input_type_list):
    """Tests get_iv method for valid input type for the 'sites' parameter"""
    _format = "json"
    site = '01491000'
    request_url = 'https://waterservices.usgs.gov/nwis/iv?format={}' \
                  '&startDT=2019-02-14&endDT=2020-02-15&sites={}'.format(_format, site)
    response_file_path = 'data/waterservices_iv.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df, md = get_iv(sites=sites, start='2019-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 563380
    assert md.url == request_url


def test_get_info(requests_mock):
    """
    Tests get_info method correctly generates the request url and returns the result in a DataFrame.
    Note that only sites and format are passed as query params
    """
    format = "rdb"
    site = '01491000%2C01645000'
    parameter_cd = "00618"
    request_url = 'https://waterservices.usgs.gov/nwis/site?sites={}&parameterCd={}&siteOutput=Expanded&format={}'.format(site, parameter_cd, format)
    response_file_path = 'data/waterservices_site.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_info(sites=["01491000", "01645000"], parameterCd="00618")
    assert type(df) is DataFrame
    assert df.size == 24
    assert md.url == request_url
    assert_metadata(requests_mock, request_url, md, site, [parameter_cd], format)


def test_get_qwdata(requests_mock):
    """Tests get_qwdata method correctly generates the request url and returns the result in a DataFrame"""
    format = "rdb"
    site = '01491000%2C01645000'
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/qwdata?site_no={}' \
                  '&qw_sample_wide=qw_sample_wide&agency_cd=USGS&format={}&pm_cd_compare=Greater+than' \
                  '&inventory_output=0&rdb_inventory_output=file&TZoutput=0&rdb_qw_attributes=expanded' \
                  '&date_format=YYYY-MM-DD&rdb_compression=value&submitted_form=brief_list'.format(site, format)
    response_file_path = 'data/waterdata_qwdata.txt'
    mock_request(requests_mock, request_url, response_file_path)
    with pytest.warns(DeprecationWarning):
        df, md = get_qwdata(sites=["01491000", "01645000"])
    assert type(df) is DataFrame
    assert df.size == 1821472
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_qwdata_site_value_types(requests_mock, site_input_type_list):
    """Tests get_qwdata method for valid input types for the 'sites' parameter"""
    _format = "rdb"
    site = '01491000'
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/qwdata?site_no={}' \
                  '&qw_sample_wide=qw_sample_wide&agency_cd=USGS&format={}&pm_cd_compare=Greater+than' \
                  '&inventory_output=0&rdb_inventory_output=file&TZoutput=0&rdb_qw_attributes=expanded' \
                  '&date_format=YYYY-MM-DD&rdb_compression=value&submitted_form=brief_list'.format(site, _format)
    response_file_path = 'data/waterdata_qwdata.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    with pytest.warns(DeprecationWarning):
        df, md = get_qwdata(sites=sites)
    assert type(df) is DataFrame
    assert df.size == 1821472


def test_get_gwlevels(requests_mock):
    """Tests get_gwlevels method correctly generates the request url and returns the result in a DataFrame."""
    format = "rdb"
    site = '434400121275801'
    request_url = 'https://waterservices.usgs.gov/nwis/gwlevels?startDT=1851-01-01' \
                  '&sites={}&format={}'.format(site, format)
    response_file_path = 'data/waterservices_gwlevels.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_gwlevels(sites=[site])
    assert type(df) is DataFrame
    assert df.size == 16
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_gwlevels_site_value_types(requests_mock, site_input_type_list):
    """Tests get_gwlevels method for valid input types for the 'sites' parameter."""
    _format = "rdb"
    site = '434400121275801'
    request_url = 'https://waterservices.usgs.gov/nwis/gwlevels?startDT=1851-01-01' \
                  '&sites={}&format={}'.format(site, _format)
    response_file_path = 'data/waterservices_gwlevels.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df, md = get_gwlevels(sites=sites)
    assert type(df) is DataFrame
    assert df.size == 16


def test_get_discharge_peaks(requests_mock):
    """Tests get_discharge_peaks method correctly generates the request url and returns the result in a DataFrame"""
    format = "rdb"
    site = '01594440'
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/peaks?format={}&site_no={}' \
                  '&begin_date=2000-02-14&end_date=2020-02-15'.format(format, site)
    response_file_path = 'data/waterservices_peaks.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_discharge_peaks(sites=[site], start='2000-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 240
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_discharge_peaks_sites_value_types(requests_mock, site_input_type_list):
    """Tests get_discharge_peaks for valid input types of the 'sites' parameter"""

    _format = "rdb"
    site = '01594440'
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/peaks?format={}&site_no={}' \
                  '&begin_date=2000-02-14&end_date=2020-02-15'.format(_format, site)
    response_file_path = 'data/waterservices_peaks.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site

    df, md = get_discharge_peaks(sites=sites, start='2000-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 240


def test_get_discharge_measurements(requests_mock):
    """Tests get_discharge_measurements method correctly generates the request url and returns the result in a
    DataFrame"""
    format = "rdb"
    site = "01594440"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/measurements?format={}&site_no={}' \
                  '&begin_date=2000-02-14&end_date=2020-02-15'.format(format, site)
    response_file_path = 'data/waterdata_measurements.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_discharge_measurements(sites=[site], start='2000-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 2130
    assert_metadata(requests_mock, request_url, md, site, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_discharge_measurements_sites_value_types(requests_mock, site_input_type_list):
    """Tests get_discharge_measurements method for valid input types for 'sites' parameter"""
    _format = "rdb"
    site = "01594440"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/measurements?format={}&site_no={}' \
                  '&begin_date=2000-02-14&end_date=2020-02-15'.format(_format, site)
    response_file_path = 'data/waterdata_measurements.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df, md = get_discharge_measurements(sites=sites, start='2000-02-14', end='2020-02-15')
    assert type(df) is DataFrame
    assert df.size == 2130


def test_get_pmcodes(requests_mock):
    """Tests get_pmcodes method correctly generates the request url and returns the result in a
    DataFrame"""
    format = "rdb"
    request_url = "https://help.waterdata.usgs.gov/code/parameter_cd_nm_query?fmt=rdb&parm_nm_cd=%2500618%25"
    response_file_path = 'data/waterdata_pmcodes.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_pmcodes(parameterCd='00618')
    assert type(df) is DataFrame
    assert df.size == 13
    assert_metadata(requests_mock, request_url, md, None, None, format)


@pytest.mark.parametrize("parameterCd_input_type_list", [True, False])
def test_get_pmcodes_parameterCd_value_types(requests_mock, parameterCd_input_type_list):
    """Tests get_pmcodes method for valid input types for the 'parameterCd' parameter"""
    _format = "rdb"
    parameterCd = '00618'
    request_url = "https://help.waterdata.usgs.gov/code/parameter_cd_nm_query?fmt={}&parm_nm_cd=%25{}%25"
    request_url = request_url.format(_format, parameterCd)
    response_file_path = 'data/waterdata_pmcodes.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if parameterCd_input_type_list:
        parameterCd = [parameterCd]
    else:
        parameterCd = parameterCd
    df, md = get_pmcodes(parameterCd=parameterCd)
    assert type(df) is DataFrame
    assert df.size == 13


def test_get_water_use_national(requests_mock):
    """Tests get_discharge_measurements method correctly generates the request url and returns the result in a
    DataFrame"""
    format = "rdb"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/water_use?rdb_compression=value&format={}&wu_year=ALL' \
                  '&wu_category=ALL&wu_county=ALL'.format(format)
    response_file_path = 'data/water_use_national.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_water_use()
    assert type(df) is DataFrame
    assert df.size == 225
    assert_metadata(requests_mock, request_url, md, None, None, format)


@pytest.mark.parametrize("year_input_type_list", [True, False])
def test_get_water_use_national_year_value_types(requests_mock, year_input_type_list):
    """Tests get_water_use method for valid input types for the 'years' parameter"""
    _format = "rdb"
    year = "ALL"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/water_use?rdb_compression=value&format={}&wu_year=ALL' \
                  '&wu_category=ALL&wu_county=ALL'.format(_format)
    response_file_path = 'data/water_use_national.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if year_input_type_list:
        years = [year]
    else:
        years = year
    df, md = get_water_use(years=years)
    assert type(df) is DataFrame
    assert df.size == 225


@pytest.mark.parametrize("county_input_type_list", [True, False])
def test_get_water_use_national_county_value_types(requests_mock, county_input_type_list):
    """Tests get_water_use method for valid input types for the 'counties' parameter"""
    _format = "rdb"
    county = "ALL"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/water_use?rdb_compression=value&format={}&wu_year=ALL' \
                  '&wu_category=ALL&wu_county=ALL'.format(_format)
    response_file_path = 'data/water_use_national.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if county_input_type_list:
        counties = [county]
    else:
        counties = county
    df, md = get_water_use(counties=counties)
    assert type(df) is DataFrame
    assert df.size == 225


@pytest.mark.parametrize("category_input_type_list", [True, False])
def test_get_water_use_national_county_value_types(requests_mock, category_input_type_list):
    """Tests get_water_use method for valid input types for the 'categories' parameter"""
    _format = "rdb"
    category = "ALL"
    request_url = 'https://nwis.waterdata.usgs.gov/nwis/water_use?rdb_compression=value&format={}&wu_year=ALL' \
                  '&wu_category=ALL&wu_county=ALL'.format(_format)
    response_file_path = 'data/water_use_national.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if category_input_type_list:
        categories = [category]
    else:
        categories = category
    df, md = get_water_use(categories=categories)
    assert type(df) is DataFrame
    assert df.size == 225


def test_get_water_use_allegheny(requests_mock):
    """Tests get_discharge_measurements method correctly generates the request url and returns the result in a
    DataFrame"""
    format = "rdb"
    request_url = 'https://nwis.waterdata.usgs.gov/PA/nwis/water_use?rdb_compression=value&format=rdb&wu_year=ALL' \
                  '&wu_category=ALL&wu_county=003&wu_area=county'
    response_file_path = 'data/water_use_allegheny.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_water_use(state="PA", counties="003")
    assert type(df) is DataFrame
    assert df.size == 1981
    assert_metadata(requests_mock, request_url, md, None, None, format)


def test_get_ratings_validation():
    """Tests get_ratings method correctly generates the request url and returns the result in a DataFrame"""
    site = "01594440"
    with pytest.raises(ValueError) as value_error:
        get_ratings(site=site, file_type="BAD")
    assert 'Unrecognized file_type: BAD, must be "base", "corr" or "exsa"' in str(value_error)


def test_get_ratings(requests_mock):
    """Tests get_ratings method correctly generates the request url and returns the result in a DataFrame"""
    format = "rdb"
    site = "01594440"
    request_url = "https://nwis.waterdata.usgs.gov/nwisweb/get_ratings/?site_no={}&file_type=base".format(site)
    response_file_path = 'data/waterservices_ratings.txt'
    mock_request(requests_mock, request_url, response_file_path)
    df, md = get_ratings(site_no=site)
    assert type(df) is DataFrame
    assert df.size == 33
    assert_metadata(requests_mock, request_url, md, site, None, format)


def test_what_sites(requests_mock):
    """Tests what_sites method correctly generates the request url and returns the result in a DataFrame"""
    format = "rdb"
    parameter_cd = '00010%2C00060'
    parameter_cd_list = ["00010","00060"]
    request_url = "https://waterservices.usgs.gov/nwis/site?bBox=-83.0%2C36.5%2C-81.0%2C38.5" \
                  "&parameterCd={}&hasDataTypeCd=dv&format={}".format(parameter_cd, format)
    response_file_path = 'data/nwis_sites.txt'
    mock_request(requests_mock, request_url, response_file_path)

    df, md = what_sites(bBox=[-83.0,36.5,-81.0,38.5], parameterCd=parameter_cd_list, hasDataTypeCd="dv")
    assert type(df) is DataFrame
    assert df.size == 2472
    assert_metadata(requests_mock, request_url, md, None, parameter_cd_list, format)


def test_get_stats(requests_mock):
    """Tests get_stats method correctly generates the request url and returns the result in a DataFrame"""
    format = "rdb"
    request_url = "https://waterservices.usgs.gov/nwis/stat?sites=01491000%2C01645000&format={}".format(format)
    response_file_path = 'data/waterservices_stats.txt'
    mock_request(requests_mock, request_url, response_file_path)

    df, md = get_stats(sites=["01491000", "01645000"])
    assert type(df) is DataFrame
    assert df.size == 51936
    assert_metadata(requests_mock, request_url, md, None, None, format)


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_stats_site_value_types(requests_mock, site_input_type_list):
    """Tests get_stats method for valid input types for the 'sites' parameter"""
    _format = "rdb"
    site = '01491000'
    request_url = "https://waterservices.usgs.gov/nwis/stat?sites={}&format={}".format(site, _format)
    response_file_path = 'data/waterservices_stats.txt'
    mock_request(requests_mock, request_url, response_file_path)
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df, md = get_stats(sites=sites)
    assert type(df) is DataFrame
    assert df.size == 51936


def mock_request(requests_mock, request_url, file_path):
    with open(file_path) as text:
        requests_mock.get(request_url, text=text.read(), headers={"mock_header": "value"})


def assert_metadata(requests_mock, request_url, md, site, parameter_cd, format):
    assert md.url == request_url
    assert isinstance(md.query_time, datetime.timedelta)
    assert md.header == {"mock_header": "value"}
    if site is not None:
        site_request_url = "https://waterservices.usgs.gov/nwis/site?sites={}&format=rdb".format(site)
        with open('data/waterservices_site.txt') as text:
            requests_mock.get(site_request_url, text=text.read())
        site_info, _ = md.site_info
        assert type(site_info) is DataFrame
    if parameter_cd is None:
        assert md.variable_info is None
    else:
        for param in parameter_cd:
            pcode_request_url = "https://help.waterdata.usgs.gov/code/parameter_cd_nm_query?fmt=rdb&parm_nm_cd=%25{}%25".format(param)
            with open('data/waterdata_pmcodes.txt') as text:
                requests_mock.get(pcode_request_url, text=text.read())
        variable_info, _ = md.variable_info
        assert type(variable_info) is DataFrame

    if format == "rdb":
        assert md.comment is not None
    else:
        assert md.comment is None



# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# Since we aren't installing package here, we mock imports of the dependencies.

# Relative paths so documentation can reference and include demos folder
import os
import sys
from importlib.metadata import version

# path to repository head
sys.path.insert(0, os.path.abspath('../..'))

# Project Information
project = 'dataretrieval'
release = version(project)
version = '.'.join(release.split('.')[:2])
author = 'Hodson et al'

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'nbsphinx',
    'nbsphinx_link',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# suffix of source documents
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# Autosummary / Automodapi settings
autosummary_generate = True
automodapi_inheritance_diagram = False
autodoc_default_options = {
    'members': True,
    'inherited-members': False,
    'private-members': True,
}

# doctest
doctest_global_setup = '''
import dataretrieval
import numpy as np
import pandas as pd
import matplotlib
'''

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_theme_options = {
    'logo_only': False,
    'display_version': True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for linkcheck -------------------------------------------

# Links to not "check" because they are problematic for the link checker
linkcheck_ignore = [
    r'https://streamstats.usgs.gov/streamstatsservices/#/',
    r'https://www.waterqualitydata.us/public_srsnames/',
    r'https://waterqualitydata.us',
    r'https://github.com/USGS-python/dataretrieval/tree/master/demos/hydroshare',
]

# Some notebooks have warnings, which nbsphinx should ignore
nbsphinx_allow_errors = True



"""Functions for downloading data from the `National Water Information System (NWIS)`_.

.. _National Water Information System (NWIS): https://waterdata.usgs.gov/nwis


.. todo::

    * Create a test to check whether functions pull multiple sites
    * Work on multi-index capabilities.
    * Check that all timezones are handled properly for each service.

"""

import re
import warnings
from io import StringIO
from typing import List, Optional, Tuple, Union

import pandas as pd
import requests

from dataretrieval.utils import BaseMetadata, format_datetime, to_str

from .utils import query

WATERDATA_BASE_URL = 'https://nwis.waterdata.usgs.gov/'
WATERDATA_URL = WATERDATA_BASE_URL + 'nwis/'
WATERSERVICE_URL = 'https://waterservices.usgs.gov/nwis/'
PARAMCODES_URL = 'https://help.waterdata.usgs.gov/code/parameter_cd_nm_query?'
ALLPARAMCODES_URL = 'https://help.waterdata.usgs.gov/code/parameter_cd_query?'

WATERSERVICES_SERVICES = ['dv', 'iv', 'site', 'stat', 'gwlevels']
WATERDATA_SERVICES = [
    'qwdata',
    'measurements',
    'peaks',
    'pmcodes',
    'water_use',
    'ratings',
]


def format_response(
    df: pd.DataFrame, service: Optional[str] = None, **kwargs
) -> pd.DataFrame:
    """Setup index for response from query.

    This function formats the response from the NWIS web services, in
    particular it sets the index of the data frame. This function tries to
    convert the NWIS response into pandas datetime values localized to UTC,
    and if possible, uses these timestamps to define the data frame index.

    Parameters
    ----------
    df: ``pandas.DataFrame``
        The data frame to format
    service: string, optional, default is None
        The NWIS service that was queried, important because the 'peaks'
        service returns a different format than the other services.
    **kwargs: optional
        Additional keyword arguments, e.g., 'multi_index'

    Returns
    -------
    df: ``pandas.DataFrame``
        The formatted data frame

    """
    mi = kwargs.pop('multi_index', True)

    if service == 'peaks':
        df = preformat_peaks_response(df)

    # check for multiple sites:
    if 'datetime' not in df.columns:
        # XXX: consider making site_no index
        return df

    elif len(df['site_no'].unique()) > 1 and mi:
        # setup multi-index
        df.set_index(['site_no', 'datetime'], inplace=True)
        if hasattr(df.index.levels[1], 'tzinfo') and df.index.levels[1].tzinfo is None:
            df = df.tz_localize('UTC', level=1)

    else:
        df.set_index(['datetime'], inplace=True)
        if hasattr(df.index, 'tzinfo') and df.index.tzinfo is None:
            df = df.tz_localize('UTC')

    return df.sort_index()


def preformat_peaks_response(df: pd.DataFrame) -> pd.DataFrame:
    """Datetime formatting for the 'peaks' service response.

    Function to format the datetime column of the 'peaks' service response.

    Parameters
    ----------
    df: ``pandas.DataFrame``
        The data frame to format

    Returns
    -------
    df: ``pandas.DataFrame``
        The formatted data frame

    """
    df['datetime'] = pd.to_datetime(df.pop('peak_dt'), errors='coerce')
    df.dropna(subset=['datetime'], inplace=True)
    return df


def get_qwdata(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    multi_index: bool = True,
    wide_format: bool = True,
    datetime_index: bool = True,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Get water sample data from qwdata service.

    .. warning::

        WARNING: Beginning in March 2024 the NWIS qw data endpoint will
        not deliver new data or updates to existing data.
        Eventually the endpoint will be retired. For updated information visit:
        https://waterdata.usgs.gov.nwis/qwdata
        For additional details, see the R package vignette:
        https://doi-usgs.github.io/dataRetrieval/articles/Status.html
        If you have additional questions about the qw data service,
        email CompTools@usgs.gov.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        If the qwdata parameter site_no is supplied, it will overwrite the
        sites parameter
    start: string, optional, default is None
        If the qwdata parameter begin_date is supplied, it will overwrite the
        start parameter (YYYY-MM-DD)
    end: string, optional, default is None
        If the qwdata parameter end_date is supplied, it will overwrite the
        end parameter (YYYY-MM-DD)
    multi_index: bool, optional
        If False, a dataframe with a single-level index (datetime) is returned,
        default is True
    wide_format : bool, optional
        If True, return data in wide format with multiple samples per row and
        one row per time, default is True
    datetime_index : bool, optional
        If True, create a datetime index, default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # get water sample information for site 11447650
        >>> df, md = dataretrieval.nwis.get_qwdata(
        ...     sites='11447650', start='2010-01-01', end='2010-02-01'
        ... )

    """
    warnings.warn(('WARNING: Starting in March 2024, the NWIS qw data endpoint is '
                       'retiring and no longer receives updates. For more information, '
                       'refer to https://waterdata.usgs.gov.nwis/qwdata and '
                       'https://doi-usgs.github.io/dataRetrieval/articles/Status.html '
                       'or email CompTools@usgs.gov.'))

    _check_sites_value_types(sites)

    kwargs['site_no'] = kwargs.pop('site_no', sites)
    kwargs['begin_date'] = kwargs.pop('begin_date', start)
    kwargs['end_date'] = kwargs.pop('end_date', end)
    kwargs['multi_index'] = multi_index
    if wide_format:
        kwargs['qw_sample_wide'] = 'qw_sample_wide'

    payload = {
        'agency_cd': 'USGS',
        'format': 'rdb',
        'pm_cd_compare': 'Greater than',
        'inventory_output': '0',
        'rdb_inventory_output': 'file',
        'TZoutput': '0',
        'rdb_qw_attributes': 'expanded',
        'date_format': 'YYYY-MM-DD',
        'rdb_compression': 'value',
        'submitted_form': 'brief_list',
    }

    # check for parameter codes, and reformat query args
    qwdata_parameter_code_field = 'parameterCd'
    if kwargs.get(qwdata_parameter_code_field):
        parameter_codes = kwargs.pop(qwdata_parameter_code_field)
        parameter_codes = to_str(parameter_codes)
        kwargs['multiple_parameter_cds'] = parameter_codes
        kwargs['param_cd_operator'] = 'OR'

        search_criteria = kwargs.get('list_of_search_criteria')
        if search_criteria:
            kwargs['list_of_search_criteria'] = '{},{}'.format(
                search_criteria, 'multiple_parameter_cds'
            )
        else:
            kwargs['list_of_search_criteria'] = 'multiple_parameter_cds'

    kwargs.update(payload)

    warnings.warn(
        'NWIS qw web services are being retired. '
        + 'See this note from the R package for more: '
        + 'https://doi-usgs.github.io/dataRetrieval/articles/qwdata_changes.html',
        category=DeprecationWarning,
    )
    response = query_waterdata('qwdata', ssl_check=ssl_check, **kwargs)

    df = _read_rdb(response.text)

    if datetime_index is True:
        df = format_datetime(df, 'sample_dt', 'sample_tm', 'sample_start_time_datum_cd')

    return format_response(df, **kwargs), NWIS_Metadata(response, **kwargs)


def get_discharge_measurements(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Get discharge measurements from the waterdata service.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        If the qwdata parameter site_no is supplied, it will overwrite the
        sites parameter
    start: string, optional, default is None
        If the qwdata parameter begin_date is supplied, it will overwrite the
        start parameter (YYYY-MM-DD)
    end: string, optional, default is None
        If the qwdata parameter end_date is supplied, it will overwrite the
        end parameter (YYYY-MM-DD)
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get discharge measurements for site 05114000
        >>> df, md = dataretrieval.nwis.get_discharge_measurements(
        ...     sites='05114000', start='2000-01-01', end='2000-01-30'
        ... )

        >>> # Get discharge measurements for sites in Alaska
        >>> df, md = dataretrieval.nwis.get_discharge_measurements(
        ...     start='2012-01-09', end='2012-01-10', stateCd='AK'
        ... )

    """
    _check_sites_value_types(sites)

    kwargs['site_no'] = kwargs.pop('site_no', sites)
    kwargs['begin_date'] = kwargs.pop('begin_date', start)
    kwargs['end_date'] = kwargs.pop('end_date', end)

    response = query_waterdata(
        'measurements', format='rdb', ssl_check=ssl_check, **kwargs
    )
    return _read_rdb(response.text), NWIS_Metadata(response, **kwargs)


def get_discharge_peaks(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    multi_index: bool = True,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Get discharge peaks from the waterdata service.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        If the waterdata parameter site_no is supplied, it will overwrite the
        sites parameter
    start: string, optional, default is None
        If the waterdata parameter begin_date is supplied, it will overwrite
        the start parameter (YYYY-MM-DD)
    end: string, optional, default is None
        If the waterdata parameter end_date is supplied, it will overwrite
        the end parameter (YYYY-MM-DD)
    multi_index: bool, optional
        If False, a dataframe with a single-level index (datetime) is returned,
        default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get discharge peaks for site 01491000
        >>> df, md = dataretrieval.nwis.get_discharge_peaks(
        ...     sites='01491000', start='1980-01-01', end='1990-01-01'
        ... )

        >>> # Get discharge peaks for sites in Hawaii
        >>> df, md = dataretrieval.nwis.get_discharge_peaks(
        ...     start='1980-01-01', end='1980-01-02', stateCd='HI'
        ... )

    """
    _check_sites_value_types(sites)

    kwargs['site_no'] = kwargs.pop('site_no', sites)
    kwargs['begin_date'] = kwargs.pop('begin_date', start)
    kwargs['end_date'] = kwargs.pop('end_date', end)
    kwargs['multi_index'] = multi_index

    response = query_waterdata('peaks', format='rdb', ssl_check=ssl_check, **kwargs)

    df = _read_rdb(response.text)

    return format_response(df, service='peaks', **kwargs), NWIS_Metadata(
        response, **kwargs
    )


def get_gwlevels(
    sites: Optional[Union[List[str], str]] = None,
    start: str = '1851-01-01',
    end: Optional[str] = None,
    multi_index: bool = True,
    datetime_index: bool = True,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Queries the groundwater level service from waterservices

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        If the waterdata parameter site_no is supplied, it will overwrite the
        sites parameter
    start: string, optional, default is '1851-01-01'
        If the waterdata parameter begin_date is supplied, it will overwrite
        the start parameter
    end: string, optional, default is None
        If the waterdata parameter end_date is supplied, it will overwrite the
        end parameter (YYYY-MM-DD)
    multi_index: bool, optional
        If False, a dataframe with a single-level index (datetime) is returned,
        default is True
    datetime_index : bool, optional
        If True, create a datetime index, default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get groundwater levels for site 434400121275801
        >>> df, md = dataretrieval.nwis.get_gwlevels(sites='434400121275801')

    """
    _check_sites_value_types(sites)

    kwargs['startDT'] = kwargs.pop('startDT', start)
    kwargs['endDT'] = kwargs.pop('endDT', end)
    kwargs['sites'] = kwargs.pop('sites', sites)
    kwargs['multi_index'] = multi_index

    response = query_waterservices('gwlevels', ssl_check=ssl_check, **kwargs)

    df = _read_rdb(response.text)

    if datetime_index is True:
        df = format_datetime(df, 'lev_dt', 'lev_tm', 'lev_tz_cd')

    return format_response(df, **kwargs), NWIS_Metadata(response, **kwargs)


def get_stats(
    sites: Optional[Union[List[str], str]] = None, ssl_check: bool = True, **kwargs
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Queries water services statistics information.

    For more information about the water services statistics service, visit
    https://waterservices.usgs.gov/docs/statistics/statistics-details/

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        USGS site number (or list of site numbers)
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Keyword Arguments
    ---------------------
    statReportType: string
        daily (default), monthly, or annual
    statTypeCd: string
        all, mean, max, min, median

    Returns
    -------
    df: ``pandas.DataFrame``
        Statistics data from the statistics service
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    .. todo::

        fix date parsing

    Examples
    --------
    .. doctest::

        >>> # Get annual water statistics for a site
        >>> df, md = dataretrieval.nwis.get_stats(
        ...     sites='01646500', statReportType='annual', statYearType='water'
        ... )

        >>> # Get monthly statistics for a site
        >>> df, md = dataretrieval.nwis.get_stats(
        ...     sites='01646500', statReportType='monthly'
        ... )

    """
    _check_sites_value_types(sites)

    response = query_waterservices(
        service='stat', sites=sites, ssl_check=ssl_check, **kwargs
    )

    return _read_rdb(response.text), NWIS_Metadata(response, **kwargs)


def query_waterdata(
    service: str, ssl_check: bool = True, **kwargs
) -> requests.models.Response:
    """
    Queries waterdata.

    Parameters
    ----------
    service: string
        Name of the service to query: 'site', 'stats', etc.
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    request: ``requests.models.Response``
        The response object from the API request to the web service
    """
    major_params = ['site_no', 'state_cd']
    bbox_params = [
        'nw_longitude_va',
        'nw_latitude_va',
        'se_longitude_va',
        'se_latitude_va',
    ]

    if not any(key in kwargs for key in major_params + bbox_params):
        raise TypeError('Query must specify a major filter: site_no, stateCd, bBox')

    elif any(key in kwargs for key in bbox_params) and not all(
        key in kwargs for key in bbox_params
    ):
        raise TypeError('One or more lat/long coordinates missing or invalid.')

    if service not in WATERDATA_SERVICES:
        raise TypeError('Service not recognized')

    url = WATERDATA_URL + service

    return query(url, payload=kwargs, ssl_check=ssl_check)


def query_waterservices(
    service: str, ssl_check: bool = True, **kwargs
) -> requests.models.Response:
    """
    Queries waterservices.usgs.gov

    For more documentation see https://waterservices.usgs.gov/docs/

    .. note::

        User must specify one major filter: sites, stateCd, or bBox

    Parameters
    ----------
    service: string
        Name of the service to query: 'site', 'stats', etc.
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Keyword Arguments
    ----------------
    bBox: string
        7-digit Hydrologic Unit Code (HUC)
    startDT: string
        Start date (e.g., '2017-12-31')
    endDT: string
        End date (e.g., '2018-01-01')
    modifiedSince: string
        Used to return only sites where attributes or period of record data
        have changed during the request period. String expected to be formatted
        in ISO-8601 duration format (e.g., 'P1D' for one day,
        'P1Y' for one year)

    Returns
    -------
    request: ``requests.models.Response``
        The response object from the API request to the web service

    """
    if not any(
        key in kwargs for key in ['sites', 'stateCd', 'bBox', 'huc', 'countyCd']
    ):
        raise TypeError(
            'Query must specify a major filter: sites, stateCd, bBox, huc, or countyCd'
        )

    if service not in WATERSERVICES_SERVICES:
        raise TypeError('Service not recognized')

    if 'format' not in kwargs:
        kwargs['format'] = 'rdb'

    url = WATERSERVICE_URL + service

    return query(url, payload=kwargs, ssl_check=ssl_check)


def get_dv(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    multi_index: bool = True,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Get daily values data from NWIS and return it as a ``pandas.DataFrame``.

    .. note:

        If no start or end date are provided, only the most recent record
        is returned.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        USGS site number (or list of site numbers)
    start: string, optional, default is None
        If the waterdata parameter startDT is supplied, it will overwrite the
        start parameter (YYYY-MM-DD)
    end: string, optional, default is None
        If the waterdata parameter endDT is supplied, it will overwrite the
        end parameter (YYYY-MM-DD)
    multi_index: bool, optional
        If True, return a multi-index dataframe, if False, return a
        single-index dataframe, default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get mean statistic daily values for site 04085427
        >>> df, md = dataretrieval.nwis.get_dv(
        ...     sites='04085427', start='2012-01-01', end='2012-06-30', statCd='00003'
        ... )

        >>> # Get the latest daily values for site 01646500
        >>> df, md = dataretrieval.nwis.get_dv(sites='01646500')

    """
    _check_sites_value_types(sites)

    kwargs['startDT'] = kwargs.pop('startDT', start)
    kwargs['endDT'] = kwargs.pop('endDT', end)
    kwargs['sites'] = kwargs.pop('sites', sites)
    kwargs['multi_index'] = multi_index

    response = query_waterservices('dv', format='json', ssl_check=ssl_check, **kwargs)
    df = _read_json(response.json())

    return format_response(df, **kwargs), NWIS_Metadata(response, **kwargs)


def get_info(ssl_check: bool = True, **kwargs) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Get site description information from NWIS.

    **Note:** *Must specify one major parameter.*

    For additional parameter options see
    https://waterservices.usgs.gov/docs/site-service/site-service-details/

    Parameters
    ----------
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Keyword Arguments
    ----------------
    sites: string or list of strings
        A list of site numbers. Sites may be prefixed with an optional agency
        code followed by a colon.
    stateCd: string
        U.S. postal service (2-digit) state code. Only 1 state can be specified
        per request.
    huc: string or list of strings
        A list of hydrologic unit codes (HUC) or aggregated watersheds. Only 1
        major HUC can be specified per request, or up to 10 minor HUCs. A major
        HUC has two digits.
    bBox: string or list of strings
        A contiguous range of decimal latitude and longitude, starting with the
        west longitude, then the south latitude, then the east longitude, and
        then the north latitude with each value separated by a comma. The
        product of the range of latitude range and longitude cannot exceed 25
        degrees. Whole or decimal degrees must be specified, up to six digits
        of precision. Minutes and seconds are not allowed.
    countyCd: string or list of strings
        A list of county numbers, in a 5 digit numeric format. The first two
        digits of a county's code are the FIPS State Code.
        (url: https://help.waterdata.usgs.gov/code/county_query?fmt=html)
    startDt: string
        Selects sites based on whether data was collected at a point in time
        beginning after startDt (start date). Dates must be in ISO-8601
        Calendar Date format (for example: 1990-01-01).
    endDt: string
        The end date for the period of record. Dates must be in ISO-8601
        Calendar Date format (for example: 1990-01-01).
    period: string
        Selects sites based on whether they were active between now
        and a time in the past. For example, period=P10W will select sites
        active in the last ten weeks.
    modifiedSince: string
        Returns only sites where site attributes or period of record data have
        changed during the request period.
    parameterCd: string or list of strings
        Returns only site data for those sites containing the requested USGS
        parameter codes.
    siteType: string or list of strings
        Restricts sites to those having one or more major and/or minor site
        types, such as stream, spring or well. For a list of all valid site
        types see https://help.waterdata.usgs.gov/site_tp_cd
        For example, siteType='ST' returns streams only.
    siteOutput: string ('basic' or 'expanded')
        Indicates the richness of metadata you want for site attributes. Note
        that for visually oriented formats like Google Map format, this
        argument has no meaning. Note: for performance reasons,
        siteOutput=expanded cannot be used if seriesCatalogOutput=true or with
        any values for outputDataTypeCd.
    seriesCatalogOutput: bool
        A switch that provides detailed period of record information for
        certain output formats. The period of record indicates date ranges for
        a certain kind of information about a site, for example the start and
        end dates for a site's daily mean streamflow.

    Returns
    -------
    df: ``pandas.DataFrame``
        Site data from the NWIS web service
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get site information for a single site
        >>> df, md = dataretrieval.nwis.get_info(sites='05114000')

        >>> # Get site information for multiple sites
        >>> df, md = dataretrieval.nwis.get_info(sites=['05114000', '09423350'])

    """
    seriesCatalogOutput = kwargs.pop('seriesCatalogOutput', None)
    if seriesCatalogOutput in ['True', 'TRUE', 'true', True]:

        warnings.warn(('WARNING: Starting in March 2024, the NWIS qw data endpoint is '
                       'retiring and no longer receives updates. For more information, '
                       'refer to https://waterdata.usgs.gov.nwis/qwdata and '
                       'https://doi-usgs.github.io/dataRetrieval/articles/Status.html '
                       'or email CompTools@usgs.gov.'))
        # convert bool to string if necessary
        kwargs['seriesCatalogOutput'] = 'True'
    else:
        # cannot have both seriesCatalogOutput and the expanded format
        kwargs['siteOutput'] = 'Expanded'

    response = query_waterservices('site', ssl_check=ssl_check, **kwargs)

    return _read_rdb(response.text), NWIS_Metadata(response, **kwargs)


def get_iv(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    multi_index: bool = True,
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """Get instantaneous values data from NWIS and return it as a DataFrame.

    .. note::

        If no start or end date are provided, only the most recent record
        is returned.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        If the waterdata parameter site_no is supplied, it will overwrite the
        sites parameter
    start: string, optional, default is None
        If the waterdata parameter startDT is supplied, it will overwrite the
        start parameter (YYYY-MM-DD)
    end: string, optional, default is None
        If the waterdata parameter endDT is supplied, it will overwrite the
        end parameter (YYYY-MM-DD)
    multi_index: bool, optional
        If False, a dataframe with a single-level index (datetime) is returned,
        default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get instantaneous discharge data for site 05114000
        >>> df, md = dataretrieval.nwis.get_iv(
        ...     sites='05114000',
        ...     start='2013-11-03',
        ...     end='2013-11-03',
        ...     parameterCd='00060',
        ... )

    """
    _check_sites_value_types(sites)

    kwargs['startDT'] = kwargs.pop('startDT', start)
    kwargs['endDT'] = kwargs.pop('endDT', end)
    kwargs['sites'] = kwargs.pop('sites', sites)
    kwargs['multi_index'] = multi_index

    response = query_waterservices(
        service='iv', format='json', ssl_check=ssl_check, **kwargs
    )

    df = _read_json(response.json())
    return format_response(df, **kwargs), NWIS_Metadata(response, **kwargs)


def get_pmcodes(
    parameterCd: Union[str, List[str]] = 'All',
    partial: bool = True,
    ssl_check: bool = True,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Return a ``pandas.DataFrame`` containing all NWIS parameter codes.

    Parameters
    ----------
    parameterCd: string or list of strings, default is 'All'
        Accepts parameter codes or names
    partial: bool, optional
        Default is True (partial querying). If False, the function will query
        only exact matches, default is True
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True

    Returns
    -------
    df: ``pandas.DataFrame``
        Data retrieved from the NWIS web service.
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get information about the '00060' pcode
        >>> df, md = dataretrieval.nwis.get_pmcodes(parameterCd='00060', partial=False)

        >>> # Get information about all 'Discharge' pcodes
        >>> df, md = dataretrieval.nwis.get_pmcodes(
        ...     parameterCd='Discharge', partial=True
        ... )

    """

    payload = {'fmt': 'rdb'}
    url = PARAMCODES_URL

    if isinstance(parameterCd, str):  # when a single code or name is given
        if parameterCd.lower() == 'all':
            payload.update({'group_cd': '%'})
            url = ALLPARAMCODES_URL
            response = query(url, payload, ssl_check=ssl_check)
            return _read_rdb(response.text), NWIS_Metadata(response)

        else:
            parameterCd = [parameterCd]

    if not isinstance(parameterCd, list):
        raise TypeError(
            'Parameter information (code or name) must be type string or list'
        )

    # Querying with a list of parameters names, codes, or mixed
    return_list = []
    for param in parameterCd:
        if isinstance(param, str):
            if partial:
                param = f'%{param}%'
            payload.update({'parm_nm_cd': param})
            response = query(url, payload, ssl_check=ssl_check)
            if len(response.text.splitlines()) < 10:  # empty query
                raise TypeError(
                    'One of the parameter codes or names entered does not'
                    'return any information, please try a different value'
                )
            return_list.append(_read_rdb(response.text))
        else:
            raise TypeError('Parameter information (code or name) must be type string')
    return pd.concat(return_list), NWIS_Metadata(response)


def get_water_use(
    years: Union[str, List[str]] = 'ALL',
    state: Optional[str] = None,
    counties: Union[str, List[str]] = 'ALL',
    categories: Union[str, List[str]] = 'ALL',
    ssl_check: bool = True,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Water use data retrieval from USGS (NWIS).

    Parameters
    ----------
    years: string or list of strings
        List or comma delimited string of years.  Must be years ending in 0 or
        5, or "ALL", which retrieves all available years, default is "ALL"
    state: string, optional, default is None
        full name, abbreviation or id
    counties: string or list of strings
        County IDs from county lookup or "ALL", default is "ALL"
    categories: string or list of strings
        List or comma delimited string of Two-letter category abbreviations,
        default is "ALL"
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True

    Returns
    -------
    df: ``pandas.DataFrame``
        Data from NWIS
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get total population for RI from the NWIS water use service
        >>> df, md = dataretrieval.nwis.get_water_use(
        ...     years='2000', state='RI', categories='TP'
        ... )

        >>> # Get the national total water use for livestock in Bgal/day
        >>> df, md = dataretrieval.nwis.get_water_use(years='2010', categories='L')

        >>> # Get 2005 domestic water use for Apache County in Arizona
        >>> df, md = dataretrieval.nwis.get_water_use(
        ...     years='2005', state='Arizona', counties='001', categories='DO'
        ... )

    """
    if years:
        if not isinstance(years, list) and not isinstance(years, str):
            raise TypeError('years must be a string or a list of strings')

    if counties:
        if not isinstance(counties, list) and not isinstance(counties, str):
            raise TypeError('counties must be a string or a list of strings')

    if categories:
        if not isinstance(categories, list) and not isinstance(categories, str):
            raise TypeError('categories must be a string or a list of strings')

    payload = {
        'rdb_compression': 'value',
        'format': 'rdb',
        'wu_year': years,
        'wu_category': categories,
        'wu_county': counties,
    }
    url = WATERDATA_URL + 'water_use'
    if state is not None:
        url = WATERDATA_BASE_URL + state + '/nwis/water_use'
        payload.update({'wu_area': 'county'})
    response = query(url, payload, ssl_check=ssl_check)
    return _read_rdb(response.text), NWIS_Metadata(response)


def get_ratings(
    site: Optional[str] = None,
    file_type: str = 'base',
    ssl_check: bool = True,
    **kwargs,
) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Rating table for an active USGS streamgage retrieval.

    Reads current rating table for an active USGS streamgage from NWISweb.
    Data is retrieved from https://waterdata.usgs.gov/nwis.

    Parameters
    ----------
    site: string, optional, default is None
        USGS site number.  This is usually an 8 digit number as a string.
        If the nwis parameter site_no is supplied, it will overwrite the site
        parameter
    file_type: string, default is "base"
        can be "base", "corr", or "exsa"
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Return
    ------
    df: ``pandas.DataFrame``
        Formatted requested data
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # Get the rating table for USGS streamgage 01594440
        >>> df, md = dataretrieval.nwis.get_ratings(site='01594440')

    """
    site = kwargs.pop('site_no', site)

    payload = {}
    url = WATERDATA_BASE_URL + 'nwisweb/get_ratings/'
    if site is not None:
        payload.update({'site_no': site})
    if file_type is not None:
        if file_type not in ['base', 'corr', 'exsa']:
            raise ValueError(
                f'Unrecognized file_type: {file_type}, must be "base", "corr" or "exsa"'
            )
        payload.update({'file_type': file_type})
    response = query(url, payload, ssl_check=ssl_check)
    return _read_rdb(response.text), NWIS_Metadata(response, site_no=site)


def what_sites(ssl_check: bool = True, **kwargs) -> Tuple[pd.DataFrame, BaseMetadata]:
    """
    Search NWIS for sites within a region with specific data.

    Parameters
    ----------
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        Accepts the same parameters as :obj:`dataretrieval.nwis.get_info`

    Return
    ------
    df: ``pandas.DataFrame``
        Formatted requested data
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    Examples
    --------
    .. doctest::

        >>> # get information about a single site
        >>> df, md = dataretrieval.nwis.what_sites(sites='05114000')

        >>> # get information about sites with phosphorus in Ohio
        >>> df, md = dataretrieval.nwis.what_sites(stateCd='OH', parameterCd='00665')

    """

    response = query_waterservices(service='site', ssl_check=ssl_check, **kwargs)

    df = _read_rdb(response.text)

    return df, NWIS_Metadata(response, **kwargs)


def get_record(
    sites: Optional[Union[List[str], str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    multi_index: bool = True,
    wide_format: bool = True,
    datetime_index: bool = True,
    state: Optional[str] = None,
    service: str = 'iv',
    ssl_check: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """
    Get data from NWIS and return it as a ``pandas.DataFrame``.

    .. note::

        If no start or end date are provided, only the most recent record is
        returned.

    Parameters
    ----------
    sites: string or list of strings, optional, default is None
        List or comma delimited string of site.
    start: string, optional, default is None
        Starting date of record (YYYY-MM-DD)
    end: string, optional, default is None
        Ending date of record. (YYYY-MM-DD)
    multi_index: bool, optional
        If False, a dataframe with a single-level index (datetime) is returned,
        default is True
    wide_format : bool, optional
        If True, return data in wide format with multiple samples per row and
        one row per time, default is True
    datetime_index : bool, optional
        If True, create a datetime index. default is True
    state: string, optional, default is None
        full name, abbreviation or id
    service: string, default is 'iv'
        - 'iv' : instantaneous data
        - 'dv' : daily mean data
        - 'qwdata' : discrete samples
        - 'site' : site description
        - 'measurements' : discharge measurements
        - 'peaks': discharge peaks
        - 'gwlevels': groundwater levels
        - 'pmcodes': get parameter codes
        - 'water_use': get water use data
        - 'ratings': get rating table
        - 'stat': get statistics
    ssl_check: bool, optional
        If True, check SSL certificates, if False, do not check SSL,
        default is True
    **kwargs: optional
        If supplied, will be used as query parameters

    Returns
    -------
        ``pandas.DataFrame`` containing requested data

    Examples
    --------
    .. doctest::

        >>> # Get latest instantaneous data from site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='iv')

        >>> # Get latest daily mean data from site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='dv')

        >>> # Get all discrete sample data from site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='qwdata')

        >>> # Get site description for site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='site')

        >>> # Get discharge measurements for site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='measurements')

        >>> # Get discharge peaks for site 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='peaks')

        >>> # Get latest groundwater level for site 434400121275801
        >>> df = dataretrieval.nwis.get_record(
        ...     sites='434400121275801', service='gwlevels'
        ... )

        >>> # Get information about the discharge parameter code
        >>> df = dataretrieval.nwis.get_record(service='pmcodes', parameterCd='00060')

        >>> # Get water use data for livestock nationally in 2010
        >>> df = dataretrieval.nwis.get_record(
        ...     service='water_use', years='2010', categories='L'
        ... )

        >>> # Get rating table for USGS streamgage 01585200
        >>> df = dataretrieval.nwis.get_record(sites='01585200', service='ratings')

        >>> # Get annual statistics for USGS station 01646500
        >>> df = dataretrieval.nwis.get_record(
        ...     sites='01646500',
        ...     service='stat',
        ...     statReportType='annual',
        ...     statYearType='water',
        ... )

    """
    _check_sites_value_types(sites)

    if service not in WATERSERVICES_SERVICES + WATERDATA_SERVICES:
        raise TypeError(f'Unrecognized service: {service}')

    if service == 'iv':
        df, _ = get_iv(
            sites=sites,
            startDT=start,
            endDT=end,
            multi_index=multi_index,
            ssl_check=ssl_check,
            **kwargs,
        )
        return df

    elif service == 'dv':
        df, _ = get_dv(
            sites=sites,
            startDT=start,
            endDT=end,
            multi_index=multi_index,
            ssl_check=ssl_check,
            **kwargs,
        )
        return df

    elif service == 'qwdata':
        df, _ = get_qwdata(
            site_no=sites,
            begin_date=start,
            end_date=end,
            multi_index=multi_index,
            wide_format=wide_format,
            ssl_check=ssl_check,
            **kwargs,
        )
        return df

    elif service == 'site':
        df, _ = get_info(sites=sites, ssl_check=ssl_check, **kwargs)
        return df

    elif service == 'measurements':
        df, _ = get_discharge_measurements(
            site_no=sites, begin_date=start, end_date=end, ssl_check=ssl_check, **kwargs
        )
        return df

    elif service == 'peaks':
        df, _ = get_discharge_peaks(
            site_no=sites,
            begin_date=start,
            end_date=end,
            multi_index=multi_index,
            ssl_check=ssl_check,
            **kwargs,
        )
        return df

    elif service == 'gwlevels':
        df, _ = get_gwlevels(
            sites=sites,
            startDT=start,
            endDT=end,
            multi_index=multi_index,
            datetime_index=datetime_index,
            ssl_check=ssl_check,
            **kwargs,
        )
        return df

    elif service == 'pmcodes':
        df, _ = get_pmcodes(ssl_check=ssl_check, **kwargs)
        return df

    elif service == 'water_use':
        df, _ = get_water_use(state=state, ssl_check=ssl_check, **kwargs)
        return df

    elif service == 'ratings':
        df, _ = get_ratings(site=sites, ssl_check=ssl_check, **kwargs)
        return df

    elif service == 'stat':
        df, _ = get_stats(sites=sites, ssl_check=ssl_check, **kwargs)
        return df

    else:
        raise TypeError(f'{service} service not yet implemented')


def _read_json(json):
    """
    Reads a NWIS Water Services formatted JSON into a ``pandas.DataFrame``.

    Parameters
    ----------
    json: dict
        A JSON dictionary response to be parsed into a ``pandas.DataFrame``

    Returns
    -------
    df: ``pandas.DataFrame``
        Times series data from the NWIS JSON
    md: :obj:`dataretrieval.utils.Metadata`
        A custom metadata object

    """
    merged_df = pd.DataFrame(columns=['site_no', 'datetime'])

    site_list = [
        ts['sourceInfo']['siteCode'][0]['value'] for ts in json['value']['timeSeries']
    ]

    # create a list of indexes for each change in site no
    # for example, [0, 21, 22] would be the first and last indeces
    index_list = [0]
    index_list.extend(
        [i + 1 for i, (a, b) in enumerate(zip(site_list[:-1], site_list[1:])) if a != b]
    )
    index_list.append(len(site_list))

    for i in range(len(index_list) - 1):
        start = index_list[i]  # [0]
        end = index_list[i + 1]  # [21]

        # grab a block containing timeseries 0:21,
        # which are all from the same site
        site_block = json['value']['timeSeries'][start:end]
        if not site_block:
            continue

        site_no = site_block[0]['sourceInfo']['siteCode'][0]['value']
        site_df = pd.DataFrame(columns=['datetime'])

        for timeseries in site_block:
            param_cd = timeseries['variable']['variableCode'][0]['value']
            # check whether min, max, mean record XXX
            option = timeseries['variable']['options']['option'][0].get('value')

            # loop through each parameter in timeseries, then concat to the merged_df
            for parameter in timeseries['values']:
                col_name = param_cd
                method = parameter['method'][0]['methodDescription']

                # if len(timeseries['values']) > 1 and method:
                if method:
                    # get method, format it, and append to column name
                    method = method.strip('[]()').lower()
                    col_name = f'{col_name}_{method}'

                if option:
                    col_name = f'{col_name}_{option}'

                record_json = parameter['value']

                if not record_json:
                    # no data in record
                    continue
                # should be able to avoid this by dumping
                record_json = str(record_json).replace("'", '"')

                # read json, converting all values to float64 and all qualifiers
                # Lists can't be hashed, thus we cannot df.merge on a list column
                record_df = pd.read_json(
                    StringIO(record_json),
                    orient='records',
                    dtype={'value': 'float64', 'qualifiers': 'unicode'},
                    convert_dates=False,
                )

                record_df['qualifiers'] = (
                    record_df['qualifiers'].str.strip('[]').str.replace("'", '')
                )

                record_df.rename(
                    columns={
                        'value': col_name,
                        'dateTime': 'datetime',
                        'qualifiers': col_name + '_cd',
                    },
                    inplace=True,
                )

                site_df = site_df.merge(record_df, how='outer', on='datetime')

        # end of site loop
        site_df['site_no'] = site_no
        merged_df = pd.concat([merged_df, site_df])

    # convert to datetime, normalizing the timezone to UTC when doing so
    if 'datetime' in merged_df.columns:
        merged_df['datetime'] = pd.to_datetime(merged_df['datetime'], utc=True)

    return merged_df


def _read_rdb(rdb):
    """
    Convert NWIS rdb table into a ``pandas.dataframe``.

    Parameters
    ----------
    rdb: string
        A string representation of an rdb table

    Returns
    -------
    df: ``pandas.dataframe``
        A formatted pandas data frame

    """
    count = 0

    for line in rdb.splitlines():
        # ignore comment lines
        if line.startswith('#'):
            count = count + 1

        else:
            break

    fields = re.split('[\t]', rdb.splitlines()[count])
    fields = [field.replace(',', '') for field in fields]
    dtypes = {
        'site_no': str,
        'dec_long_va': float,
        'dec_lat_va': float,
        'parm_cd': str,
        'parameter_cd': str,
    }

    df = pd.read_csv(
        StringIO(rdb),
        delimiter='\t',
        skiprows=count + 2,
        names=fields,
        na_values='NaN',
        dtype=dtypes,
    )

    df = format_response(df)
    return df


def _check_sites_value_types(sites):
    if sites:
        if not isinstance(sites, list) and not isinstance(sites, str):
            raise TypeError('sites must be a string or a list of strings')


class NWIS_Metadata(BaseMetadata):
    """Metadata class for NWIS service, derived from BaseMetadata.

    Attributes
    ----------
    url : str
        Response url
    query_time: datetme.timedelta
        Response elapsed time
    header: requests.structures.CaseInsensitiveDict
        Response headers
    comments: str | None
        Metadata comments, if any
    site_info: tuple[pd.DataFrame, NWIS_Metadata] | None
        Site information if the query included `site_no`, `sites`, `stateCd`,
        `huc`, `countyCd` or `bBox`. `site_no` is preferred over `sites` if
        both are present.
    variable_info: tuple[pd.DataFrame, NWIS_Metadata] | None
        Variable information if the query included `parameterCd`.

    """

    def __init__(self, response, **parameters) -> None:
        """Generates a standard set of metadata informed by the response with specific
        metadata for NWIS data.

        Parameters
        ----------
        response: Response
            Response object from requests module
        parameters: unpacked dictionary
            Unpacked dictionary of the parameters supplied in the request

        Returns
        -------
        md: :obj:`dataretrieval.nwis.NWIS_Metadata`
            A ``dataretrieval`` custom :obj:`dataretrieval.nwis.NWIS_Metadata` object.

        """
        super().__init__(response)

        comments = ''
        for line in response.text.splitlines():
            if line.startswith('#'):
                comments += line.lstrip('#') + '\n'
        if comments:
            self.comment = comments

        self._parameters = parameters

    @property
    def site_info(self) -> Optional[Tuple[pd.DataFrame, BaseMetadata]]:
        """
        Return
        ------
        df: ``pandas.DataFrame``
            Formatted requested data from calling `nwis.what_sites`
        md: :obj:`dataretrieval.nwis.NWIS_Metadata`
            A NWIS_Metadata object
        """
        if 'site_no' in self._parameters:
            return what_sites(sites=self._parameters['site_no'])

        elif 'sites' in self._parameters:
            return what_sites(sites=self._parameters['sites'])

        elif 'stateCd' in self._parameters:
            return what_sites(stateCd=self._parameters['stateCd'])

        elif 'huc' in self._parameters:
            return what_sites(huc=self._parameters['huc'])

        elif 'countyCd' in self._parameters:
            return what_sites(countyCd=self._parameters['countyCd'])

        elif 'bBox' in self._parameters:
            return what_sites(bBox=self._parameters['bBox'])

        else:
            return None  # don't set metadata site_info attribute

    @property
    def variable_info(self) -> Optional[Tuple[pd.DataFrame, BaseMetadata]]:
        # define variable_info metadata based on parameterCd if available
        if 'parameterCd' in self._parameters:
            return get_pmcodes(parameterCd=self._parameters['parameterCd'])



from typing import Dict, List, Union

import pandas as pd
import requests

ResponseFormat = 'json'  # json, xml

# WaterWatch won't receive any new features but it will continue to operate.
waterwatch_url = 'https://waterwatch.usgs.gov/webservices/'


def _read_json(data: Dict) -> pd.DataFrame:
    return pd.DataFrame(data).T


def get_flood_stage(
    sites: List[str] = None, fmt: str = 'DF'
) -> Union[pd.DataFrame, Dict]:
    """
    Retrieves flood stages for a list of station numbers.

    Parameters
    ----------
    sites: List of strings
        Site numbers
    fmt: ``pandas.DataFrame`` or dict
        Returned format: Default is "DF" for ``pandas.DataFrame``, else
        a dictionary is returned.

    Returns
    -------
    station_stages: ``pandas.Dataframe`` or dict
        contains station numbers and their flood stages.
        If no flood stage for a station, ``None`` is returned.

    Examples
    --------
    .. doctest::

        >> stations = ["07144100", "07144101"]
        >> res = get_flood_stage(stations, fmt="dict")  # dictionary output
        >> print(res)
        {'07144100': {'action_stage': '20',
                      'flood_stage': '22',
                      'moderate_flood_stage': '25',
                      'major_flood_stage': '26'},
         '07144101': None}
        >> print(get_flood_stage(stations))
        >> print(res)
                action_stage flood_stage moderate_flood_stage major_flood_stage
        07144100           20          22                   25                26
        07144101         None        None                 None              None
        50057000           16          20                   24                30

    """
    res = requests.get(waterwatch_url + 'floodstage', params={'format': ResponseFormat})

    if res.ok:
        json_res = res.json()
        stages = {
            site['site_no']: {k: v for k, v in site.items() if k != 'site_no'}
            for site in json_res['sites']
        }
    else:
        raise requests.RequestException(f'[{res.status_code}] - {res.reason}')

    if not sites:
        stations_stages = stages
    else:
        stations_stages = {}
        for site in sites:
            try:
                stations_stages[site] = stages[site]
            except KeyError:
                stations_stages[site] = None

    if fmt == 'dict':
        return stations_stages
    else:
        return _read_json(stations_stages)



"""List of 2-digit state codes with commented full names."""
state_codes = [
    'al',  # Alabama
    'ak',  # Alaska
    'az',  # Arizona
    'ar',  # Arkansas
    'ca',  # California
    'co',  # Colorado
    'ct',  # Connecticut
    'de',  # Delaware
    'dc',  # District of Columbia
    'fl',  # Florida
    'ga',  # Georgia
    'hi',  # Hawaii
    'id',  # Idaho
    'il',  # Illinois
    'in',  # Indiana
    'ia',  # Iowa
    'ks',  # Kansas
    'ky',  # Kentucky
    'la',  # Louisiana
    'me',  # Maine
    'md',  # Maryland
    'ma',  # Massachusetts
    'mi',  # Michigan
    'mn',  # Minnesota
    'ms',  # Mississippi
    'mo',  # Missouri
    'mt',  # Montana
    'ne',  # Nebraska
    'nv',  # Nevada
    'nh',  # New Hampshire
    'nj',  # New Jersey
    'nm',  # New Mexico
    'ny',  # New York
    'nc',  # North Carolina
    'nd',  # North Dakota
    'oh',  # Ohio
    'ok',  # Oklahoma
    'or',  # Oregon
    'pa',  # Pennsylvania
    'ri',  # Rhode Island
    'sc',  # South Carolina
    'sd',  # South Dakota
    'tn',  # Tennessee
    'tx',  # Texas
    'ut',  # Utah
    'vt',  # Vermont
    'va',  # Virginia
    'wa',  # Washington
    'wv',  # West Virginia
    'wi',  # Wisconsin
    'wy',  # Wyoming
]



import datetime
import unittest.mock as mock

import numpy as np
import pandas as pd
import pytest

from dataretrieval.nwis import NWIS_Metadata
from dataretrieval.nwis import get_info, get_record, preformat_peaks_response, get_iv, what_sites

START_DATE = '2018-01-24'
END_DATE = '2018-01-25'

DATETIME_COL = 'datetime'
SITENO_COL = 'site_no'


def test_measurements_service():
    """Test measurement service
    """
    start = '2018-01-24'
    end = '2018-01-25'
    service = 'measurements'
    site = '03339000'
    df = get_record(site, start, end, service=service)
    return df


def test_measurements_service_answer():
    df = test_measurements_service()
    # check parsing
    assert df.iloc[0]['measurement_nu'] == 801


def test_iv_service():
    """Unit test of instantaneous value service
    """
    start = START_DATE
    end = END_DATE
    service = 'iv'
    site = ['03339000', '05447500', '03346500']
    return get_record(site, start, end, service=service)


def test_iv_service_answer():
    df = test_iv_service()
    # check multiindex function
    assert df.index.names == [SITENO_COL, DATETIME_COL], "iv service returned incorrect index: {}".format(df.index.names)


def test_preformat_peaks_response():
    # make a data frame with a "peak_dt" datetime column
    # it will have some nan and none values
    data = {"peak_dt": ["2000-03-22",
                        np.nan,
                        None],
            "peak_va": [1000,
                        2000,
                        3000]
    }
    # turn data into dataframe
    df = pd.DataFrame(data)
    # run preformat function
    df = preformat_peaks_response(df)
    # assertions
    assert 'datetime' in df.columns
    assert df['datetime'].isna().sum() == 0


@pytest.mark.parametrize("site_input_type_list", [True, False])
def test_get_record_site_value_types(site_input_type_list):
    """Test that get_record method for valid input types for the 'sites' parameter."""
    start = '2018-01-24'
    end = '2018-01-25'
    service = 'measurements'
    site = '03339000'
    if site_input_type_list:
        sites = [site]
    else:
        sites = site
    df = get_record(sites=sites, start=start, end=end, service=service)
    assert df.iloc[0]['measurement_nu'] == 801


if __name__ == '__main__':
    test_measurements_service_answer()
    test_iv_service_answer()


# tests using real queries to USGS webservices
# these specific queries represent some edge-cases and the tests to address
# incomplete date-time information

def test_inc_date_01():
    """Test based on GitHub Issue #47 - lack of timestamp for measurement."""
    site = "403451073585601"
    # make call expecting a warning to be thrown due to incomplete dates
    with pytest.warns(UserWarning):
        df = get_record(site, "1980-01-01", "1990-01-01", service='gwlevels')
    # assert that there are indeed incomplete dates
    assert any(pd.isna(df.index) == True)
    # assert that the datetime index is there
    assert df.index.name == 'datetime'
    # make call without defining a datetime index and check that it isn't there
    df2 = get_record(site, "1980-01-01", "1990-01-01", service='gwlevels',
                     datetime_index=False)
    # assert shape of both dataframes is the same (contain the same data)
    assert df.shape == df2.shape
    # assert that the datetime index is not there
    assert df2.index.name != 'datetime'


def test_inc_date_02():
    """Test based on GitHub Issue #47 - lack of month, day, or time."""
    site = "180049066381200"
    # make call expecting a warning to be thrown due to incomplete dates
    with pytest.warns(UserWarning):
        df = get_record(site, "1900-01-01", "2013-01-01", service='gwlevels')
    # assert that there are indeed incomplete dates
    assert any(pd.isna(df.index) == True)
    # assert that the datetime index is there
    assert df.index.name == 'datetime'
    # make call without defining a datetime index and check that it isn't there
    df2 = get_record(site, "1900-01-01", "2013-01-01", service='gwlevels',
                     datetime_index=False)
    # assert shape of both dataframes is the same (contain the same data)
    assert df.shape == df2.shape
    # assert that the datetime index is not there
    assert df2.index.name != 'datetime'


def test_inc_date_03():
    """Test based on GitHub Issue #47 - lack of day, and times."""
    site = "290000095192602"
    # make call expecting a warning to be thrown due to incomplete dates
    with pytest.warns(UserWarning):
        df = get_record(site, "1975-01-01", "2000-01-01", service='gwlevels')
    # assert that there are indeed incomplete dates
    assert any(pd.isna(df.index) == True)
    # assert that the datetime index is there
    assert df.index.name == 'datetime'
    # make call without defining a datetime index and check that it isn't there
    df2 = get_record(site, "1975-01-01", "2000-01-01", service='gwlevels',
                     datetime_index=False)
    # assert shape of both dataframes is the same (contain the same data)
    assert df.shape == df2.shape
    # assert that the datetime index is not there
    assert df2.index.name != 'datetime'


class TestTZ:
    """Tests relating to GitHub Issue #60."""
    sites, _ = what_sites(stateCd='MD')

    def test_multiple_tz_01(self):
        """Test based on GitHub Issue #60 - error merging different time zones."""
        # this test fails before issue #60 is fixed
        iv, _ = get_iv(sites=self.sites.site_no.values[:25].tolist())
        # assert that the datetime column exists
        assert 'datetime' in iv.index.names
        # assert that it is a datetime type
        assert isinstance(iv.index[0][1], datetime.datetime)

    def test_multiple_tz_02(self):
        """Test based on GitHub Issue #60 - confirm behavior for same tz."""
        # this test passes before issue #60 is fixed
        iv, _ = get_iv(sites=self.sites.site_no.values[:20].tolist())
        # assert that the datetime column exists
        assert 'datetime' in iv.index.names
        # assert that it is a datetime type
        assert isinstance(iv.index[0][1], datetime.datetime)


class TestSiteseriesCatalogOutput:
    """Tests relating to GitHub Issue #34."""

    def test_seriesCatalogOutput_get_record(self):
        """Test setting seriesCatalogOutput to true with get_record."""
        data = get_record(huc='20', parameterCd='00060',
                          service='site', seriesCatalogOutput='True')
        # assert that expected data columns are present
        assert 'begin_date' in data.columns
        assert 'end_date' in data.columns
        assert 'count_nu' in data.columns

    def test_seriesCatalogOutput_get_info(self):
        """Test setting seriesCatalogOutput to true with get_info."""
        data, _ = get_info(
            huc='20', parameterCd='00060', seriesCatalogOutput='TRUE')
        # assert that expected data columns are present
        assert 'begin_date' in data.columns
        assert 'end_date' in data.columns
        assert 'count_nu' in data.columns

    def test_seriesCatalogOutput_bool(self):
        """Test setting seriesCatalogOutput with a boolean."""
        data, _ = get_info(
            huc='20', parameterCd='00060', seriesCatalogOutput=True)
        # assert that expected data columns are present
        assert 'begin_date' in data.columns
        assert 'end_date' in data.columns
        assert 'count_nu' in data.columns

    def test_expandedrdb_get_record(self):
        """Test default expanded_rdb format with get_record."""
        data = get_record(huc='20', parameterCd='00060',
                          service='site', seriesCatalogOutput='False')
        # assert that seriesCatalogOutput columns are not present
        assert 'begin_date' not in data.columns
        assert 'end_date' not in data.columns
        assert 'count_nu' not in data.columns

    def test_expandedrdb_get_info(self):
        """Test default expanded_rdb format with get_info."""
        data, _ = get_info(huc='20', parameterCd='00060')
        # assert that seriesCatalogOutput columns are not present
        assert 'begin_date' not in data.columns
        assert 'end_date' not in data.columns
        assert 'count_nu' not in data.columns


def test_empty_timeseries():
    """Test based on empty case from GitHub Issue #26."""
    df = get_record(sites='011277906', service='iv',
                    start='2010-07-20', end='2010-07-20')
    assert df.empty is True


class TestMetaData:
    """Tests of NWIS metadata setting, 

    Notes
    -----

    - Originally based on GitHub Issue #73.
    - Modified to site_info and variable_info as properties, not callables.
    """

    def test_set_metadata_info_site(self):
        """Test metadata info is set when site parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, sites='01491000')
        # assert that site_info is implemented
        assert md.site_info
        
    def test_set_metadata_info_site_no(self):
        """Test metadata info is set when site_no parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, site_no='01491000')
        # assert that site_info is implemented
        assert md.site_info

    def test_set_metadata_info_stateCd(self):
        """Test metadata info is set when stateCd parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, stateCd='RI')
        # assert that site_info is implemented
        assert md.site_info

    def test_set_metadata_info_huc(self):
        """Test metadata info is set when huc parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, huc='01')
        # assert that site_info is implemented
        assert md.site_info

    def test_set_metadata_info_bbox(self):
        """Test metadata info is set when bbox parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, bBox='-92.8,44.2,-88.9,46.0')
        # assert that site_info is implemented
        assert md.site_info

    def test_set_metadata_info_countyCd(self):
        """Test metadata info is set when countyCd parameter is supplied."""
        # mock the query response
        response = mock.MagicMock()
        # make metadata call
        md = NWIS_Metadata(response, countyCd='01001')
        # assert that site_info is implemented
        assert md.site_info
